# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.utility_objs.extra_monitor_scp_messages\
    import (
        SetRouterTimeoutMessage)


class SetRouterTimeoutProcess(AbstractMultiConnectionProcess):
    """
    How to send messages to set router timeouts. These messages need to be
    sent to cores running the extra monitor binary.

    .. note::
        SCAMP sets wait2 to zero by default!

    .. note::
        Timeouts are specified in a weird floating point format.
        See the SpiNNaker datasheet for details.
    """

    def set_wait1_timeout(self, mantissa, exponent, core_subsets):
        """
        The wait1 timeout is the time from when a packet is received to
        when emergency routing becomes enabled.

        :param int mantissa: Timeout mantissa (0 to 15)
        :param int exponent: Timeout exponent (0 to 15)
        :param ~spinn_machine.CoreSubsets core_subsets:
            Where the extra monitors that manage the routers are located.
        """
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._set_timeout(
                    core_subset, processor_id, mantissa, exponent, wait=1)

    def set_wait2_timeout(self, mantissa, exponent, core_subsets):
        """
        The wait2 timeout is the time from when a packet has emergency
        routing enabled for it to when it is dropped.

        :param int mantissa: Timeout mantissa (0 to 15)
        :param int exponent: Timeout exponent (0 to 15)
        :param ~spinn_machine.CoreSubsets core_subsets:
            Where the extra monitors that manage the routers are located.
        """
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._set_timeout(
                    core_subset, processor_id, mantissa, exponent, wait=2)

    def _set_timeout(self, core, processor_id, mantissa, exponent, wait):
        """
        Set a timeout for a router controlled by an extra monitor on a core.
        This is not a parallelised operation in order to aid debugging when
        it fails.

        :param ~spinn_machine.CoreSubset core:
        :param int processor_id:
        :param int mantissa:
        :param int exponent:
        :param int wait:
        """
        self._send_request(SetRouterTimeoutMessage(
            core.x, core.y, processor_id,
            timeout_mantissa=mantissa, timeout_exponent=exponent, wait=wait))
        self._finish()
        self.check_for_error()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from pacman.operations.router_compressors.routing_compression_checker import (
    codify_table, compare_route)
from pacman.utilities.algorithm_utilities.routes_format import format_route
from spinn_front_end_common.data import FecDataView
logger = FormatAdapter(logging.getLogger(__name__))


def generate_routing_compression_checker_report(
        routing_tables, compressed_routing_tables):
    """
    Make a full report of how the compressed covers all routes in the
    and uncompressed routing table.

    :param str report_folder: the folder to store the resulting report
    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        the original routing tables
    :param compressed_routing_tables: the compressed routing tables
    :type compressed_routing_tables:
        ~pacman.model.routing_tables.MulticastRoutingTables
    """
    file_name = os.path.join(
        FecDataView.get_run_dir_path(),
        "routing_compression_checker_report.rpt")

    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(
                routing_tables.routing_tables,
                "Generating routing compression checker report")
            f.write("If this table did not raise an exception compression "
                    "was fully checked.\n\n")
            f.write("The format is:\n"
                    "Chip x, y\n"
                    "\t Uncompressed Route\n"
                    "\t\tCompressed Route\n\n")

            for original in progress.over(routing_tables.routing_tables):
                x = original.x
                y = original.y
                f.write(f"Chip: X:{x} Y:{y}\n")

                compressed_table = compressed_routing_tables.\
                    get_routing_table_for_chip(x, y)
                compressed_dict = codify_table(compressed_table)
                for o_route in original.multicast_routing_entries:
                    f.write(f"\t{format_route(o_route)}\n")
                    compare_route(o_route, compressed_dict, f=f)
    except IOError:
        logger.exception(
            "Generate router comparison reports: "
            "Can't open file {} for writing.", file_name)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import logging
import os
from spinn_utilities.config_holder import (get_config_int, get_config_str)
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import (
    FecTimer, GlobalProvenance, TimerCategory)
from spinn_front_end_common.utility_models import ChipPowerMonitorMachineVertex
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.interface_functions.compute_energy_used\
    import (JOULES_PER_SPIKE, MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD,
            MILLIWATTS_PER_FRAME_ACTIVE_COST, MILLIWATTS_PER_FPGA,
            MILLIWATTS_PER_IDLE_CHIP)
from spinn_machine.machine import Machine

logger = FormatAdapter(logging.getLogger(__name__))


class EnergyReport(object):
    """
    This class creates a report about the approximate total energy
    consumed by a SpiNNaker job execution.
    """

    __slots__ = ()

    #: converter between joules to kilowatt hours
    JOULES_TO_KILOWATT_HOURS = 3600000

    # energy report file name
    _DETAILED_FILENAME = "detailed_energy_report.rpt"
    _SUMMARY_FILENAME = "summary_energy_report.rpt"

    def write_energy_report(self, power_used):
        """
        Writes the report.

        :param ~spinn_machine.Machine machine: the machine
        :param PowerUsed power_used:
        """
        report_dir = FecDataView.get_run_dir_path()

        # detailed report path
        detailed_report = os.path.join(report_dir, self._DETAILED_FILENAME)

        # summary report path
        summary_report = os.path.join(report_dir, self._SUMMARY_FILENAME)

        # create detailed report
        with open(detailed_report, "w", encoding="utf-8") as f:
            self._write_detailed_report(power_used, f)

        # create summary report
        with open(summary_report, "w", encoding="utf-8") as f:
            self._write_summary_report(f, power_used)

    @classmethod
    def _write_summary_report(cls, f, power_used):
        """
        Write summary file.

        :param ~io.TextIOBase f: file writer
        :param PowerUsed power_used:
        """
        # pylint: disable=too-many-arguments, too-many-locals

        # figure runtime in milliseconds with time scale factor
        runtime_total_ms = (
                FecDataView.get_current_run_time_ms() *
                FecDataView.get_time_scale_factor())

        # write summary data
        f.write("Summary energy file\n-------------------\n\n")
        f.write(
            "Energy used by chips during runtime is "
            f"{power_used.chip_energy_joules} Joules "
            f"{cls.__report_time(runtime_total_ms / 1000)}\n")
        f.write(
            f"Energy used by FPGAs is {power_used.fpga_total_energy_joules} "
            "Joules over the entire time the machine was booted "
            f"{cls.__report_time(power_used.booted_time_secs)}\n")
        f.write(
            f"Energy used by FPGAs is {power_used.fpga_exec_energy_joules} "
            "Joules over the runtime period "
            f"{cls.__report_time(runtime_total_ms / 1000)}\n")
        f.write(
            "Energy used by outside router / cooling during the runtime "
            f"period is {power_used.baseline_joules} Joules\n")
        f.write(
            "Energy used by packet transmissions is "
            f"{power_used.packet_joules} Joules "
            f"{cls.__report_time(power_used.total_time_secs)}\n")
        f.write(
            "Energy used during the mapping process is "
            f"{power_used.mapping_joules} Joules "
            f"{cls.__report_time(power_used.mapping_time_secs)}\n")
        f.write(
            "Energy used by the data generation process is "
            f"{power_used.data_gen_joules} Joules "
            f"{cls.__report_time(power_used.data_gen_time_secs)}\n")
        f.write(
            "Energy used during the loading process is "
            f"{power_used.loading_joules} Joules "
            f"{cls.__report_time(power_used.loading_time_secs)}\n")
        f.write(
            "Energy used during the data extraction process is "
            f"{power_used.saving_joules} Joules "
            f"{cls.__report_time(power_used.saving_time_secs)}\n")
        f.write(
            "Total energy used by the simulation over {} milliseconds is:\n"
            "     {} Joules, or\n"
            "     {} estimated average Watts, or\n"
            "     {} kWh\n".format(
                power_used.total_time_secs * 1000,
                power_used.total_energy_joules,
                power_used.total_energy_joules / power_used.total_time_secs,
                power_used.total_energy_joules /
                cls.JOULES_TO_KILOWATT_HOURS))

    @staticmethod
    def __report_time(time):
        """
        :param float time:
        :rtype: str
        """
        if time < 1:
            return f"(over {time * 1000} milliseconds)"
        else:
            return f"(over {time} seconds)"

    def _write_detailed_report(self, power_used, f):
        """
        Write detailed report and calculate costs.

        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        """
        runtime_total_ms = (
                FecDataView.get_current_run_time_ms() *
                FecDataView.get_time_scale_factor())

        # write warning about accuracy etc
        self._write_warning(f)

        # figure out packet cost
        f.write(f"The packet cost is {power_used.packet_joules} Joules\n")

        # figure FPGA cost over all booted and during runtime cost
        self._write_fpga_cost(power_used, f)

        # figure load time cost
        self._write_load_time_cost(power_used, f)

        # figure extraction time cost
        self._write_data_extraction_time_cost(power_used, f)

        # sort what to report by chip
        active_chips = defaultdict(dict)
        for placement in FecDataView.iterate_placements_by_vertex_type(
                ChipPowerMonitorMachineVertex):
            labels = active_chips[placement.x, placement.y]
            labels[placement.p] = placement.vertex.label
        for xy in active_chips:
            self._write_chips_active_cost(
                xy, active_chips[xy], runtime_total_ms, power_used, f)

    def _write_warning(self, f):
        """
        Writes the warning about this being only an estimate.

        :param ~io.TextIOBase f: the writer
        """
        f.write(
            "This report is based off energy estimates for individual "
            "components of the SpiNNaker machine. It is not meant to be "
            "completely accurate. But does use provenance data gathered from "
            "the machine to estimate the energy usage and therefore should "
            "be in the right ballpark.\n\n\n")
        f.write(
            "The energy components we use are as follows:\n\n"
            "The energy usage for a chip when all cores are 100% active for "
            f"a millisecond is {MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD} Joules.\n"
            "The energy usage for a chip when all cores are not active for a "
            f"millisecond is {MILLIWATTS_PER_IDLE_CHIP} Joules.\n"
            "The energy used by the machine for firing a packet is "
            f"{JOULES_PER_SPIKE} Joules.\n"
            "The energy used by each active FPGA per millisecond is "
            f"{MILLIWATTS_PER_FPGA} Joules.\n\n\n")

    def _write_fpga_cost(self, power_used, f):
        """
        FPGA cost model calculation.

        :param PowerUsed power_used: the runtime
        :param ~io.TextIOBase f: the file writer
        """
        version = get_config_int("Machine", "version")
        # if not spalloc, then could be any type of board
        if (not get_config_str("Machine", "spalloc_server") and
                not get_config_str("Machine", "remote_spinnaker_url")):
            # if a spinn2 or spinn3 (4 chip boards) then they have no fpgas
            if version in (2, 3):
                f.write(
                    f"A SpiNN-{version} board does not contain any FPGA's, "
                    f"and so its energy cost is 0 \n")
                return
            elif version not in (4, 5):
                # no idea where we are; version unrecognised
                raise ConfigurationException(
                    "Do not know what the FPGA setup is for this version of "
                    "SpiNNaker machine.")

            # if a spinn4 or spinn5 board, need to verify if wrap-arounds
            # are there, if not then assume fpgas are turned off.
            if power_used.num_fpgas == 0:
                # no active fpgas
                f.write(
                    f"The FPGA's on the SpiNN-{version} board are turned off "
                    f"and therefore the energy used by the FPGA is 0\n")
                return
            # active fpgas; fall through to shared main part report

        # print out as needed for spalloc and non-spalloc versions
        if version is None:
            f.write(
                f"{power_used.num_fpgas} FPGAs on the Spalloc-ed boards are "
                "turned on and therefore the energy used by the FPGA during "
                "the entire time the machine was booted (which was "
                f"{power_used.total_time_secs * 1000} ms) is "
                f"{power_used.fpga_total_energy_joules} Joules. "
                "The usage during execution was "
                f"{power_used.fpga_exec_energy_joules} Joules.")
        else:
            f.write(
                f"{power_used.num_fpgas} FPGA's on the SpiNN-{version} board "
                "are turned on and therefore the energy used by the FPGA "
                "during the entire time the machine was booted (which was "
                f"{power_used.total_time_secs * 1000} ms) is "
                f"{power_used.fpga_total_energy_joules} Joules. "
                "The usage during execution was "
                f"{power_used.fpga_exec_energy_joules} Joules.")

    @staticmethod
    def _write_chips_active_cost(
            chip_coord, labels, runtime_total_ms, power_used, f):
        """
        Figure out the chip active cost during simulation.

        :param tuple(int,int) chip_coord: the x,y of the chip to consider
        :param dict(int,str) labels: vertex labels for the active cores
        :param float runtime_total_ms:
        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        :return: energy cost
        """
        (x, y) = chip_coord
        f.write("\n")

        # detailed report print out
        for core in range(Machine.DEFAULT_MAX_CORES_PER_CHIP):
            if core in labels:
                label = f" (running {labels[core]})"
            else:
                label = ""
            energy = power_used.get_core_active_energy_joules(x, y, core)
            f.write(
                f"processor {x}:{y}:{core}{label} used {energy} Joules by "
                "being active during the execution of the simulation\n")

        # TAKE INTO ACCOUNT IDLE COST
        idle_cost = runtime_total_ms * MILLIWATTS_PER_IDLE_CHIP

        f.write(
            f"The chip at {x},{y} used {idle_cost} Joules by "
            "being idle during the execution of the simulation\n")

    @staticmethod
    def _write_load_time_cost(power_used, f):
        """
        Energy usage from the loading phase.

        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        """
        # find time in milliseconds
        with GlobalProvenance() as db:
            total_time_ms = db.get_timer_sum_by_category(TimerCategory.LOADING)

        # handle active routers etc
        active_router_cost = (
            power_used.loading_time_secs * 1000 * power_used.num_frames *
            MILLIWATTS_PER_FRAME_ACTIVE_COST)

        # detailed report write
        f.write(
            "The amount of time used during the loading process is "
            f"{total_time_ms} milliseconds.\nAssumed only 2 monitor cores is "
            "executing that this point. We also assume that there is a "
            "baseline active router/cooling component that is using "
            f"{active_router_cost} Joules. Overall the energy usage is "
            f"{power_used.loading_joules} Joules.\n")

    @staticmethod
    def _write_data_extraction_time_cost(power_used, f):
        """
        Data extraction cost.

        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        """
        # find time
        with GlobalProvenance() as db:
            total_time_ms = db.get_timer_sum_by_algorithm(
                FecTimer.APPLICATION_RUNNER)

        # handle active routers etc
        energy_cost_of_active_router = (
            total_time_ms * power_used.num_frames *
            MILLIWATTS_PER_FRAME_ACTIVE_COST)

        # detailed report
        f.write(
            "The amount of time used during the data extraction process is "
            f"{total_time_ms} milliseconds.\nAssumed only 2 monitor cores is "
            "executing at this point. We also assume that there is a baseline "
            "active router/cooling component that is using "
            f"{energy_cost_of_active_router} Joules. Hence the overall energy "
            f"usage is {power_used.saving_joules} Joules.\n")



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.data import FecDataView
from .profile_data import ProfileData

_PROFILE_HEADER_SIZE_BYTES = BYTES_PER_WORD
_SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES = 2 * BYTES_PER_WORD
_BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION = BYTES_PER_WORD


def get_profile_region_size(n_samples):
    """
    Get the size of the region of the profile data.

    :param int n_samples: number of different samples to record
    :return: the size in bytes used by the profile region
    :rtype: int
    """
    return _PROFILE_HEADER_SIZE_BYTES + (
        n_samples * _SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES)


def reserve_profile_region(spec, region, n_samples):
    """
    Reserves the profile region for recording the profile data.

    :param ~data_specification.DataSpecificationGenerator spec:
        the DSG specification writer
    :param int region: region ID for the profile data
    :param int n_samples: number of elements being sampled
    """
    size = get_profile_region_size(n_samples)
    spec.reserve_memory_region(
        region=region, size=size, label="profilerRegion")


def write_profile_region_data(spec, region, n_samples):
    """
    Writes the profile region data.

    :param ~data_specification.DataSpecificationGenerator spec:
        the DSG specification writer
    :param int region: region ID for the profile data
    :param int n_samples: number of elements being sampled
    """
    spec.switch_write_focus(region)
    spec.write_value(n_samples)


def get_profiling_data(profile_region, tag_labels, placement):
    """
    Utility function to get profile data from a profile region.

    :param int profile_region: DSG region to get profiling data out of SDRAM
    :param list(str) tag_labels: labels for the profiling data
    :param ~pacman.model.placements.Placement placement: placement
    :rtype: ProfileData
    """
    txrx = FecDataView.get_transceiver()
    profile_data = ProfileData(tag_labels)

    address = locate_memory_region_for_placement(
        placement=placement, region=profile_region)

    # Read the profiling data size
    words_written = txrx.read_word(placement.x, placement.y, address)

    # Read the profiling data
    if words_written != 0:
        address += _BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION
        profile_data.add_data(txrx.read_memory(
            placement.x, placement.y, address, words_written * BYTES_PER_WORD))

    return profile_data



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3
import time
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB

_DDL_FILE = os.path.join(os.path.dirname(__file__),
                         "db.sql")
_SECONDS_TO_MICRO_SECONDS_CONVERSION = 1000
#: Name of the database in the data folder


def _timestamp():
    return int(time.time() * _SECONDS_TO_MICRO_SECONDS_CONVERSION)


class BaseDatabase(SQLiteDB, AbstractContextManager):
    """
    Specific implementation of the Database for SQLite 3.

    There should only ever be a single Database Object in use at any time.
    In the case of application_graph_changed the first should closed and
    a new one created.

    If 2 database objects where opened with the database_file they hold the
    same data. Unless someone else deletes that file.

    .. note::
        *Not thread safe on the same database file!*
        Threads can access different DBs just fine.
    """

    __slots__ = ["_database_file"]

    def __init__(self, database_file=None, *, read_only=False,
                 row_factory=sqlite3.Row, text_factory=memoryview):
        """
        :param str database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data.
            If omitted the default location will be used.
        """
        if database_file:
            self._database_file = database_file
        else:
            self._database_file = self.default_database_file()
        super().__init__(
            self._database_file, read_only=read_only, row_factory=row_factory,
            text_factory=text_factory, ddl_file=_DDL_FILE)

    @classmethod
    def default_database_file(cls):
        return os.path.join(FecDataView.get_run_dir_path(),
                            f"data{FecDataView.get_reset_str()}.sqlite3")

    def _get_core_id(self, cursor, x, y, p):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :rtype: int
        """
        for row in cursor.execute(
                """
                SELECT core_id FROM core
                WHERE x = ? AND y = ? AND processor = ?
                LIMIT 1
                """, (x, y, p)):
            return row["core_id"]
        cursor.execute(
            """
            INSERT INTO core(x, y, processor) VALUES(?, ?, ?)
            """, (x, y, p))
        return cursor.lastrowid



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class StreamingContextManager(object):
    """
    The implementation of the context manager object for streaming
    configuration control.
    """
    __slots__ = ["_gatherers"]

    def __init__(self, gatherers):
        """
        :param iterable(DataSpeedUpPacketGatherMachineVertex) gatherers:
        """
        self._gatherers = list(gatherers)

    def __enter__(self):
        for gatherer in self._gatherers:
            gatherer.load_system_routing_tables()
        for gatherer in self._gatherers:
            gatherer.set_cores_for_data_streaming()

    def __exit__(self, _type, _value, _tb):
        for gatherer in self._gatherers:
            gatherer.unset_cores_for_data_streaming()
        for gatherer in self._gatherers:
            gatherer.load_application_routing_tables()
        return False



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_generates_data_specification import (
    AbstractGeneratesDataSpecification)
from .abstract_has_associated_binary import AbstractHasAssociatedBinary
from .abstract_machine_allocation_controller import (
    AbstractMachineAllocationController)
from .abstract_rewrites_data_specification import (
    AbstractRewritesDataSpecification)
from .abstract_send_me_multicast_commands_vertex import (
    AbstractSendMeMulticastCommandsVertex)
from .abstract_vertex_with_dependent_vertices import (
    AbstractVertexWithEdgeToDependentVertices)
from .abstract_supports_database_injection import (
    AbstractSupportsDatabaseInjection)
from .abstract_supports_bit_field_generation import (
    AbstractSupportsBitFieldGeneration)
from .abstract_supports_bit_field_routing_compression import (
    AbstractSupportsBitFieldRoutingCompression)
from .abstract_can_reset import AbstractCanReset
from .has_custom_atom_key_map import HasCustomAtomKeyMap

__all__ = ["AbstractGeneratesDataSpecification",
           "AbstractHasAssociatedBinary",
           "AbstractMachineAllocationController",
           "AbstractRewritesDataSpecification",
           "AbstractSendMeMulticastCommandsVertex",
           "AbstractSupportsDatabaseInjection",
           "AbstractVertexWithEdgeToDependentVertices", "AbstractCanReset",
           "AbstractSupportsBitFieldGeneration",
           "AbstractSupportsBitFieldRoutingCompression",
           "HasCustomAtomKeyMap"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_machine import Router
from pacman import exceptions
from pacman.model.graphs.machine import (
    MachineFPGAVertex, MachineSpiNNakerLinkVertex)
from pacman.utilities.algorithm_utilities.routing_algorithm_utilities import (
    get_app_partitions)
from pacman.utilities.algorithm_utilities.routes_format import format_route
from spinn_front_end_common.data import FecDataView
from .router_summary import RouterSummary

logger = FormatAdapter(logging.getLogger(__name__))

_LINK_LABELS = {0: 'E', 1: 'NE', 2: 'N', 3: 'W', 4: 'SW', 5: 'S'}

_C_ROUTING_TABLE_DIR = "compressed_routing_tables_generated"
_COMPARED_FILENAME = "comparison_of_compressed_uncompressed_routing_tables.rpt"
_COMPRESSED_ROUTING_SUMMARY_FILENAME = "compressed_routing_summary.rpt"
_PARTITIONING_FILENAME = "partitioned_by_vertex.rpt"
_PLACEMENT_VTX_GRAPH_FILENAME = "placement_by_vertex_using_graph.rpt"
_PLACEMENT_VTX_SIMPLE_FILENAME = "placement_by_vertex_without_graph.rpt"
_PLACEMENT_CORE_GRAPH_FILENAME = "placement_by_core_using_graph.rpt"
_PLACEMENT_CORE_SIMPLE_FILENAME = "placement_by_core_without_graph.rpt"
_ROUTING_FILENAME = "edge_routing_info.rpt"
_ROUTING_SUMMARY_FILENAME = "routing_summary.rpt"
_ROUTING_TABLE_DIR = "routing_tables_generated"
_SDRAM_FILENAME = "chip_sdram_usage_by_core.rpt"
_TAGS_FILENAME = "tags.rpt"
_VIRTKEY_FILENAME = "virtual_key_space_information_report.rpt"

_LOWER_16_BITS = 0xFFFF


def tag_allocator_report():
    """
    Reports the tags that are being used by the tool chain for this
    simulation.
    """
    tag_infos = FecDataView.get_tags()
    file_name = os.path.join(FecDataView.get_run_dir_path(), _TAGS_FILENAME)
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(
                len(list(tag_infos.ip_tags)) +
                len(list(tag_infos.reverse_ip_tags)),
                "Reporting Tags")
            for ip_tag in progress.over(tag_infos.ip_tags, False):
                f.write(str(ip_tag) + "\n")
            for reverse_ip_tag in progress.over(tag_infos.reverse_ip_tags):
                f.write(str(reverse_ip_tag) + "\n")
    except IOError:
        logger.error(
            "Generate tag report: Can't open file {} for writing.", file_name)


def placer_reports_with_application_graph():
    """
    Reports that can be produced from placement given a application
    graph's existence.
    """
    placement_report_with_application_graph_by_vertex()
    placement_report_with_application_graph_by_core()


def router_summary_report():
    """
    Generates a text file of routing summaries.

    :rtype: RouterSummary
    """
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _ROUTING_SUMMARY_FILENAME)
    progress = ProgressBar(FecDataView.get_machine().n_chips,
                           "Generating Routing summary report")
    routing_tables = FecDataView.get_uncompressed()
    return _do_router_summary_report(file_name, progress, routing_tables)


def router_compressed_summary_report(routing_tables):
    """
    Generates a text file of routing summaries.

    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        The in-operation COMPRESSED routing tables.
    :rtype: RouterSummary
    """
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _COMPRESSED_ROUTING_SUMMARY_FILENAME)
    progress = ProgressBar(FecDataView.get_machine().n_chips,
                           "Generating Routing summary report")
    return _do_router_summary_report(file_name, progress, routing_tables)


def _do_router_summary_report(file_name, progress, routing_tables):
    """
    :param str file_name:
    :param ~spinn_utilities.progress_bar.Progress progress:
    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        The compressed or uncompressed tables being reported
    :return: RouterSummary
    """
    time_date_string = time.strftime("%c")
    convert = Router.convert_routing_table_entry_to_spinnaker_route
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            f.write("        Routing Summary Report\n")
            f.write("        ======================\n\n")
            f.write(f"Generated: {time_date_string} "
                    f"for target machine '{FecDataView.get_ipaddress()}'\n\n")

            total_entries = 0
            max_entries = 0
            max_none_defaultable = 0
            max_link_only = 0
            max_spinnaker_routes = 0
            for (x, y) in progress.over(
                    FecDataView.get_machine().chip_coordinates):
                table = routing_tables.get_routing_table_for_chip(x, y)
                if table is not None:
                    entries = table.number_of_entries
                    defaultable = table.number_of_defaultable_entries
                    link_only = 0
                    spinnaker_routes = set()
                    for entry in table.multicast_routing_entries:
                        if not entry.processor_ids:
                            link_only += 1
                        spinnaker_routes.add(convert(entry))
                    f.write(
                        f"Chip {x}:{y} has {entries} entries of which "
                        f"{defaultable} are defaultable and {link_only} link "
                        f"only with {len(spinnaker_routes)} unique spinnaker "
                        "routes\n")
                    total_entries += entries
                    max_entries = max(max_entries, entries)
                    max_none_defaultable = max(
                        max_none_defaultable, entries - defaultable)
                    max_link_only = max(max_link_only, link_only)
                    max_spinnaker_routes = max(
                        max_spinnaker_routes, len(spinnaker_routes))

            f.write(
                f"\nTotal entries {total_entries}, max per chip {max_entries} "
                f"max non-defaultable {max_none_defaultable} "
                f"max link only {max_link_only} "
                f"max unique spinnaker routes {max_spinnaker_routes}\n\n")
            return RouterSummary(
                total_entries, max_entries, max_none_defaultable,
                max_link_only, max_spinnaker_routes)

    except IOError:
        logger.exception(
            "Generate routing summary report: Can't open file {} for writing.",
            file_name)


def router_report_from_paths():
    """
    Generates a text file of routing paths.
    """
    file_name = os.path.join(FecDataView.get_run_dir_path(), _ROUTING_FILENAME)
    time_date_string = time.strftime("%c")
    partitions = get_app_partitions()
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(len(partitions),
                                   "Generating Routing path report")

            f.write("        Edge Routing Report\n")
            f.write("        ===================\n\n")
            f.write(f"Generated: {time_date_string} "
                    f"for target machine '{FecDataView.get_ipaddress()}'\n\n")

            for partition in progress.over(partitions):
                _write_one_router_partition_report(f, partition)
    except IOError:
        logger.exception(
            "Generate routing reports: Can't open file {} for writing.",
            file_name)


def _write_one_router_partition_report(f, partition):
    """
    :param ~io.FileIO f:
    :param AbstractSingleSourcePartition partition:
    """
    source = partition.pre_vertex
    outgoing = source.splitter.get_out_going_vertices(partition.identifier)
    f.write(f"Source Application vertex {source}, partition"
            f" {partition.identifier}\n")

    routing_infos = FecDataView.get_routing_infos()
    for edge in partition.edges:
        for m_vertex in outgoing:
            source_placement = FecDataView.get_placement_of_vertex(m_vertex)
            r_info = routing_infos.get_routing_info_from_pre_vertex(
                m_vertex, partition.identifier)
            path = _search_route(source_placement, r_info.key_and_mask)
            f.write(
                f"    Edge '{edge.label}', "
                f"from vertex: '{edge.pre_vertex.label}' "
                f"to vertex: '{edge.post_vertex.label}'{path}\n")

            # End one entry:
            f.write("\n")


def partitioner_report():
    """
    Generate report on the partitioning of vertices.
    """
    # Cycle through all vertices, and for each cycle through its vertices.
    # For each vertex, describe its core mapping.
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _PARTITIONING_FILENAME)
    time_date_string = time.strftime("%c")
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(FecDataView.get_n_vertices(),
                                   "Generating partitioner report")

            f.write("        Partitioning Information by Vertex\n")
            f.write("        ===============================\n\n")
            f.write(f"Generated: {time_date_string} for target machine "
                    f"'{FecDataView.get_ipaddress()}'\n\n")

            for vertex in progress.over(FecDataView.iterate_vertices()):
                _write_one_vertex_partition(f, vertex)
    except IOError:
        logger.exception(
            "Generate partitioning reports: Can't open file {} for writing.",
            file_name)


def _write_one_vertex_partition(f, vertex):
    """
    :param ~io.FileIO f:
    :param ~pacman.model.graphs.application.ApplicationVertex vertex:
    """
    vertex_name = vertex.label
    vertex_model = vertex.__class__.__name__
    num_atoms = vertex.n_atoms
    f.write(f"**** Vertex: '{vertex_name}'\n")
    f.write(f"Model: {vertex_model}\n")
    f.write(f"Pop size: {num_atoms}\n")
    f.write("Machine Vertices:\n")

    # Sort by slice and then by label
    machine_vertices = sorted(vertex.machine_vertices,
                              key=lambda x: x.label)
    machine_vertices = sorted(machine_vertices,
                              key=lambda x: x.vertex_slice.lo_atom)
    for sv in machine_vertices:
        f.write(f"  Slice {sv.vertex_slice}\n")
    f.write("\n")


def placement_report_with_application_graph_by_vertex():
    """
    Generate report on the placement of vertices onto cores by vertex.
    """
    # Cycle through all vertices, and for each cycle through its vertices.
    # For each vertex, describe its core mapping.
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _PLACEMENT_VTX_GRAPH_FILENAME)
    time_date_string = time.strftime("%c")
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(FecDataView.get_n_vertices(),
                                   "Generating placement report")

            f.write("        Placement Information by Vertex\n")
            f.write("        ===============================\n\n")
            f.write(f"Generated: {time_date_string} "
                    f"for target machine '{FecDataView.get_ipaddress()}'\n\n")

            for vertex in progress.over(FecDataView.iterate_vertices()):
                _write_one_vertex_application_placement(f, vertex)
    except IOError:
        logger.exception(
            "Generate placement reports: Can't open file {} for writing.",
            file_name)


def _write_one_vertex_application_placement(f, vertex):
    """
    :param ~io.FileIO f:
    :param ~pacman.model.graphs.application.ApplicationVertex vertex:
    """
    vertex_name = vertex.label
    vertex_model = vertex.__class__.__name__
    num_atoms = vertex.n_atoms
    f.write(f"**** Vertex: '{vertex_name}'\n")
    f.write(f"Model: {vertex_model}\n")
    f.write(f"Pop size: {num_atoms}\n")
    f.write("Machine Vertices: \n")

    # Sort by slice and then by label
    machine_vertices = sorted(vertex.machine_vertices,
                              key=lambda vert: vert.label)
    machine_vertices = sorted(machine_vertices,
                              key=lambda vert: vert.vertex_slice.lo_atom)
    for sv in machine_vertices:
        if isinstance(sv, MachineSpiNNakerLinkVertex):
            f.write("  Slice {} on SpiNNaker Link {}, board {},"
                    " linked to chip {}\n"
                    .format(sv.vertex_slice, sv.spinnaker_link_id,
                            sv.board_address, sv.linked_chip_coordinates))
        elif isinstance(sv, MachineFPGAVertex):
            f.write("  Slice {} on FGPA {}, FPGA link {}, board {},"
                    " linked to chip {}\n"
                    .format(sv.vertex_slice, sv.fpga_id, sv.fpga_link_id,
                            sv.board_address, sv.linked_chip_coordinates))
        else:
            cur_placement = FecDataView.get_placement_of_vertex(sv)
            x, y, p = cur_placement.x, cur_placement.y, cur_placement.p
            f.write(f"  Slice {sv.vertex_slice} on core ({x}, {y}, {p}) \n")
    f.write("\n")


def placement_report_with_application_graph_by_core():
    """
    Generate report on the placement of vertices onto cores by core.
    """
    # File 2: Placement by core.
    # Cycle through all chips and by all cores within each chip.
    # For each core, display what is held on it.
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _PLACEMENT_CORE_GRAPH_FILENAME)
    time_date_string = time.strftime("%c")
    try:
        machine = FecDataView.get_machine()
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(machine.n_chips,
                                   "Generating placement by core report")

            f.write("        Placement Information by Core\n")
            f.write("        =============================\n\n")
            f.write(f"Generated: {time_date_string} "
                    f"for target machine '{FecDataView.get_ipaddress()}'\n\n")

            for chip in progress.over(machine.chips):
                _write_one_chip_application_placement(f, chip)
    except IOError:
        logger.exception(
            "Generate_placement_reports: Can't open file {} for writing.",
            file_name)


def _write_one_chip_application_placement(f, chip):
    """
    :param ~io.FileIO f:
    :param ~spinn_machine.Chip chip:
    :param ~pacman.model.placements.Placements placements:
    """
    written_header = False
    total_sdram = None
    for placement in FecDataView.iterate_placements_on_core(chip.x, chip.y):
        if not written_header:
            f.write(f"**** Chip: ({chip.x}, {chip.y})\n")
            f.write(f"Application cores: {len(list(chip.processors))}\n")
            written_header = True
        pro_id = placement.p
        vertex = placement.vertex
        app_vertex = vertex.app_vertex
        if app_vertex is not None:
            vertex_label = app_vertex.label
            vertex_model = app_vertex.__class__.__name__
            vertex_atoms = app_vertex.n_atoms
            f.write("  Processor {}: Vertex: '{}', pop size: {}\n".format(
                pro_id, vertex_label, vertex_atoms))
            f.write("              Slice on this core: {}\n".format(
                vertex.vertex_slice))
            f.write("              Model: {}\n".format(vertex_model))
        else:
            f.write("  Processor {}: System Vertex: '{}'\n".format(
                pro_id, vertex.label))
            f.write("              Model: {}\n".format(
                vertex.__class__.__name__))

        sdram = vertex.sdram_required
        f.write("              SDRAM required: {}; {} per timestep\n\n"
                .format(sdram.fixed, sdram.per_timestep))
        if total_sdram is None:
            total_sdram = sdram
        else:
            total_sdram += sdram

    if total_sdram is not None:
        f.write("Total SDRAM on chip ({} available): {}; {} per-timestep\n\n"
                .format(chip.sdram.size, total_sdram.fixed,
                        total_sdram.per_timestep))


def sdram_usage_report_per_chip():
    """
    Reports the SDRAM used per chip.
    """
    file_name = os.path.join(FecDataView.get_run_dir_path(), _SDRAM_FILENAME)
    n_placements = FecDataView.get_n_placements()
    time_date_string = time.strftime("%c")
    progress = ProgressBar(
        (n_placements * 2 + FecDataView.get_machine().n_chips * 2),
        "Generating SDRAM usage report")
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            f.write("        Memory Usage by Core\n")
            f.write("        ====================\n\n")
            f.write(f"Generated: {time_date_string} "
                    f"for target machine '{FecDataView.get_ipaddress()}'\n\n")
            f.write("Planned by partitioner\n")
            f.write("----------------------\n")
            _sdram_usage_report_per_chip_with_timesteps(
                f, FecDataView.get_plan_n_timestep(), progress, False, False)
            f.write("\nActual space reserved on the machine\n")
            f.write("----------------------\n")
            _sdram_usage_report_per_chip_with_timesteps(
                f, FecDataView.get_max_run_time_steps(), progress, True, True)
    except IOError:
        logger.exception("Generate_placement_reports: Can't open file {} for "
                         "writing.", file_name)


def _sdram_usage_report_per_chip_with_timesteps(
        f, timesteps, progress, end_progress, details):
    """
    :param ~io.FileIO f:
    :param int timesteps: Either the plan or data timesteps
        depending on which is being reported
    :param ~spinn_utilities.progress_bar.ProgressBar progress:
    :param bool end_progress:
    :param bool details: If True will get costs printed by regions
    """
    f.write(f"Based on {timesteps} timesteps\n\n")
    used_sdram_by_chip = dict()
    placements = sorted(
        FecDataView.iterate_placemements(), key=lambda x: x.vertex.label)
    for placement in progress.over(placements, False):
        vertex_sdram = placement.vertex.sdram_required
        core_sdram = vertex_sdram.get_total_sdram(timesteps)
        x, y, p = placement.x, placement.y, placement.p
        if details:
            vertex_sdram.report(
                timesteps=timesteps,
                preamble=f"core ({x},{y},{p})", target=f)
        else:
            f.write(
                "SDRAM reqs for core ({},{},{}) is {} KB ({} bytes) for {}\n"
                "".format(x, y, p, int(core_sdram / 1024.0), core_sdram,
                          placement))
        key = (x, y)
        if key not in used_sdram_by_chip:
            used_sdram_by_chip[key] = core_sdram
        else:
            used_sdram_by_chip[key] += core_sdram
    for chip in progress.over(FecDataView.get_machine().chips, end_progress):
        try:
            used_sdram = used_sdram_by_chip[chip.x, chip.y]
            if used_sdram:
                f.write(
                    "**** Chip: ({}, {}) has total memory usage of"
                    " {} KB ({} bytes) out of a max of "
                    "{} KB ({} bytes)\n\n".format(
                        chip.x, chip.y,
                        int(used_sdram / 1024.0), used_sdram,
                        int(chip.sdram.size / 1024.0), chip.sdram.size))
        except KeyError:
            # Do Nothing
            pass


def routing_info_report(extra_allocations):
    """
    Generates a report which says which keys is being allocated to each
    vertex.
    """
    file_name = os.path.join(FecDataView.get_run_dir_path(), _VIRTKEY_FILENAME)
    routing_infos = FecDataView.get_routing_infos()
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            vertex_partitions = set(
                (p.pre_vertex, p.identifier)
                for p in FecDataView.iterate_partitions())
            vertex_partitions.update(extra_allocations)
            progress = ProgressBar(len(vertex_partitions),
                                   "Generating Routing info report")
            for pre_vert, part_id in progress.over(vertex_partitions):
                _write_vertex_virtual_keys(f, pre_vert, part_id, routing_infos)
            progress.end()
    except IOError:
        logger.exception("generate virtual key space information report: "
                         "Can't open file {} for writing.", file_name)


def _write_vertex_virtual_keys(f, pre_vertex, part_id, routing_infos):
    """
    :param ~io.FileIO f:
    :param ~pacman.model.graphs.application.ApplicationVertex pre_vertex:
    :param str part_id:
    :param ~pacman.model.routing_info.RoutingInfo routing_infos:
    :param ~spinn_utilities.progress_bar.ProgressBar progress:
    """
    rinfo = routing_infos.get_routing_info_from_pre_vertex(
        pre_vertex, part_id)
    # Might be None if the partition has no outgoing vertices e.g. a Poisson
    # source replaced by SDRAM comms
    if rinfo is not None:
        f.write(f"Vertex: {pre_vertex}\n")
        f.write("    Partition: {}, Routing Info: {}\n".format(
            part_id, rinfo.key_and_mask))
        for m_vertex in pre_vertex.splitter.get_out_going_vertices(part_id):
            r_info = routing_infos.get_routing_info_from_pre_vertex(
                m_vertex, part_id)
            if r_info is not None:
                f.write("    Machine Vertex: {}, Slice: {}, Routing Info: {}\n"
                        .format(m_vertex, m_vertex.vertex_slice,
                                r_info.key_and_mask))


def router_report_from_router_tables():
    """
    Report the uncompressed routing tables.
    """
    top_level_folder = os.path.join(
        FecDataView.get_run_dir_path(), _ROUTING_TABLE_DIR)
    routing_tables = FecDataView.get_uncompressed().routing_tables
    if not os.path.exists(top_level_folder):
        os.mkdir(top_level_folder)
    progress = ProgressBar(routing_tables, "Generating Router table report")
    for routing_table in progress.over(routing_tables):
        if routing_table.number_of_entries:
            generate_routing_table(routing_table, top_level_folder)


def router_report_from_compressed_router_tables(routing_tables):
    """
    Report the compressed routing tables.

    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        the compressed routing tables
    """
    top_level_folder = os.path.join(
        FecDataView.get_run_dir_path(), _C_ROUTING_TABLE_DIR)
    if not os.path.exists(top_level_folder):
        os.mkdir(top_level_folder)
    progress = ProgressBar(routing_tables.routing_tables,
                           "Generating compressed router table report")
    for routing_table in progress.over(routing_tables.routing_tables):
        if routing_table.number_of_entries:
            generate_routing_table(routing_table, top_level_folder)


def generate_routing_table(routing_table, top_level_folder):
    """
    :param routing_table: The routing table to describe
    :type routing_table:
        ~pacman.model.routing_tables.AbstractMulticastRoutingTable
    :param str top_level_folder:
    """
    file_name = "routing_table_{}_{}.rpt".format(
        routing_table.x, routing_table.y)
    file_path = os.path.join(top_level_folder, file_name)
    try:
        with open(file_path, "w", encoding="utf-8") as f:
            f.write("Router contains {} entries\n".format(
                routing_table.number_of_entries))

            f.write("{: <5s} {: <10s} {: <10s} {: <10s} {: <7s} {}\n".format(
                "Index", "Key", "Mask", "Route", "Default", "[Cores][Links]"))
            f.write(
                "{:-<5s} {:-<10s} {:-<10s} {:-<10s} {:-<7s} {:-<14s}\n".format(
                    "", "", "", "", "", ""))
            line_format = "{: >5d} {}\n"

            entry_count = 0
            n_defaultable = 0
            for entry in routing_table.multicast_routing_entries:
                index = entry_count & _LOWER_16_BITS
                entry_str = line_format.format(index, format_route(entry))
                entry_count += 1
                if entry.defaultable:
                    n_defaultable += 1
                f.write(entry_str)
            f.write(f"{n_defaultable} Defaultable entries\n")
    except IOError:
        logger.exception("Generate_placement_reports: Can't open file"
                         " {} for writing.", file_path)


def _compression_ratio(uncompressed, compressed):
    """
    Get the compression ratio, as a percentage.

    :param int uncompressed:
    :param int compressed:
    :rtype: float
    """
    if uncompressed == 0:
        return 0
    return (uncompressed - compressed) / float(uncompressed) * 100


def generate_comparison_router_report(compressed_routing_tables):
    """
    Make a report on comparison of the compressed and uncompressed
    routing tables.

    :param compressed_routing_tables: the compressed routing tables
    :type compressed_routing_tables:
        ~pacman.model.routing_tables.MulticastRoutingTables
    """
    routing_tables = FecDataView.get_uncompressed().routing_tables
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), _COMPARED_FILENAME)
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            progress = ProgressBar(
                routing_tables,
                "Generating comparison of router table report")
            total_uncompressed = 0
            total_compressed = 0
            max_compressed = 0
            uncompressed_for_max = 0
            for table in progress.over(routing_tables):
                x = table.x
                y = table.y
                compressed_table = compressed_routing_tables.\
                    get_routing_table_for_chip(x, y)
                n_entries_uncompressed = table.number_of_entries
                total_uncompressed += n_entries_uncompressed
                n_entries_compressed = compressed_table.number_of_entries
                total_compressed += n_entries_compressed
                ratio = _compression_ratio(
                    n_entries_uncompressed, n_entries_compressed)
                f.write(
                    f"Uncompressed table at {x}:{y} has "
                    f"{n_entries_uncompressed} entries whereas compressed "
                    f"table has {n_entries_compressed} entries. This is a "
                    f"decrease of {ratio}%\n")
                if max_compressed < n_entries_compressed:
                    max_compressed = n_entries_compressed
                    uncompressed_for_max = n_entries_uncompressed
            if total_uncompressed > 0:
                ratio = _compression_ratio(
                    total_uncompressed, total_compressed)
                f.write(
                    f"\nTotal has {total_uncompressed} entries whereas "
                    f"compressed tables have {total_compressed} entries. "
                    f"This is an average decrease of {ratio}%\n")
                ratio = _compression_ratio(
                    uncompressed_for_max, max_compressed)
                f.write(
                    f"Worst case has {uncompressed_for_max} entries whereas "
                    f"compressed tables have {max_compressed} entries. This "
                    f"is a decrease of {ratio}%\n")
    except IOError:
        logger.exception(
            "Generate router comparison reports: "
            "Can't open file {} for writing.", file_name)


def _search_route(source_placement, key_and_mask):
    """
    :param ~pacman.model.placements.Placement source_placement:
    :param ~pacman.model.routing_info.BaseKeyAndMask key_and_mask:
    :rtype: tuple(str, int)
    """
    # Create text for starting point
    machine = FecDataView.get_machine()
    source_vertex = source_placement.vertex
    text = ""
    if isinstance(source_vertex, MachineSpiNNakerLinkVertex):
        text = "        Virtual SpiNNaker Link on {}:{}:{} -> ".format(
            source_placement.x, source_placement.y, source_placement.p)
        slink = machine.get_spinnaker_link_with_id(
            source_vertex.spinnaker_link_id)
        x = slink.connected_chip_x
        y = slink.connected_chip_y
    elif isinstance(source_vertex, MachineFPGAVertex):
        text = "        Virtual FPGA Link on {}:{}:{} -> ".format(
            source_placement.x, source_placement.y, source_placement.p)
        flink = machine.get_fpga_link_with_id(
            source_vertex.fpga_id, source_vertex.fpga_link_id)
        x = flink.connected_chip_x
        y = flink.connected_chip_y
    else:
        x = source_placement.x
        y = source_placement.y
        text = "        {}:{}:{} -> ".format(
            source_placement.x, source_placement.y, source_placement.p)

    # If the destination is virtual, replace with the real destination chip
    text += _recursive_trace_to_destinations(
        x, y, key_and_mask, pre_space="        ")
    return text


# Locates the destinations of a route
def _recursive_trace_to_destinations(
        chip_x, chip_y, key_and_mask, pre_space):
    """
    Recursively search though routing tables till no more entries are
    registered with this key

    :param int chip_x:
    :param int chip_y
    :param ~pacman.model.routing_info.BaseKeyAndMask key_and_mask:
    :rtype: str
    """
    chip = FecDataView.get_chip_at(chip_x, chip_y)

    text = f"-> Chip {chip_x}:{chip_y}"
    routing_tables = FecDataView.get_uncompressed()
    table = routing_tables.get_routing_table_for_chip(chip_x, chip_y)
    entry = _locate_routing_entry(table, key_and_mask.key)
    new_pre_space = pre_space + (" " * len(text))
    first = True
    for link_id in entry.link_ids:
        if not first:
            text += f"\n{pre_space}"
        link = chip.router.get_link(link_id)
        text += f"-> {link}"
        text += _recursive_trace_to_destinations(
            link.destination_x, link.destination_y, key_and_mask,
            new_pre_space)
        first = False

    return text


def _locate_routing_entry(current_router, key):
    """
    Locate the entry from the router based off the edge

    :param ~spinn_machine.MulticastRoutingTable current_router:
        the current router being used in the trace
    :param int key: the key being used by the source placement
    :return: the routing table entry
    :rtype: ~spinn_machine.MulticastRoutingEntry
    :raise PacmanRoutingException:
        when there is no entry located on this router.
    """
    for entry in current_router.multicast_routing_entries:
        if entry.mask & key == entry.routing_entry_key:
            return entry
    raise exceptions.PacmanRoutingException("no entry located")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.routing_tables import MulticastRoutingTables
from pacman.model.routing_tables.compressed_multicast_routing_table import (
    CompressedMulticastRoutingTable)
from spinn_front_end_common.data import FecDataView


def read_routing_tables_from_machine():
    """
    Reads compressed routing tables from a SpiNNaker machine.

    :rtype: ~pacman.model.routing_tables.MulticastRoutingTables
    """
    routing_tables = FecDataView.get_uncompressed()
    progress = ProgressBar(
        routing_tables, "Reading Routing Tables from Machine")
    machine_routing_tables = MulticastRoutingTables()
    app_id = FecDataView.get_app_id()
    transceiver = FecDataView.get_transceiver()
    for routing_table in progress.over(routing_tables):
        # get multicast entries from machine
        machine_routing_table = _read_routing_table(
            transceiver, routing_table, app_id)
        machine_routing_tables.add_routing_table(machine_routing_table)

    return machine_routing_tables


def _read_routing_table(transceiver, table, app_id):
    """
    :param ~spinnman.transceiver.Transceiver transceiver:
    :param ~.UnCompressedMulticastRoutingTable table:
    :param int app_id:
    """
    machine_routing_table = CompressedMulticastRoutingTable(table.x, table.y)
    for routing_entry in transceiver.get_multicast_routes(
            table.x, table.y, app_id):
        machine_routing_table.add_multicast_routing_entry(routing_entry)
    return machine_routing_table



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import struct
import numpy
from enum import IntEnum
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.messages.eieio import EIEIOPrefix, EIEIOType
from spinnman.messages.eieio.data_messages import EIEIODataHeader
from pacman.model.resources import ReverseIPtagResource, VariableSDRAM
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.utilities.utility_calls import get_field_based_keys
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    SendsBuffersFromHostPreBufferedImpl, AbstractReceiveBuffersToHost)
from spinn_front_end_common.interface.buffer_management.storage_objects\
    .buffered_sending_region import (
        get_n_bytes)
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import (
        BufferedSendingRegion)
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
    AbstractSupportsDatabaseInjection)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.buffer_management.recording_utilities \
    import (get_recording_header_array, get_recording_header_size,
            get_recording_data_constant_size)
from spinn_front_end_common.utilities.utility_objs import ExecutableType

logger = FormatAdapter(logging.getLogger(__name__))

_DEFAULT_MALLOC_REGIONS = 2
_ONE_WORD = struct.Struct("<I")
_TWO_SHORTS = struct.Struct("<HH")

# The microseconds per timestep will be divided by this for the max offset
_MAX_OFFSET_DENOMINATOR = 10
# The max offset modulo to stop spikes in simple cases moving to the next ts
_MAX_OFFSET_MODULO = 1000


class ReverseIPTagMulticastSourceMachineVertex(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary, AbstractSupportsDatabaseInjection,
        ProvidesProvenanceDataFromMachineImpl,
        SendsBuffersFromHostPreBufferedImpl, AbstractReceiveBuffersToHost):
    """
    A model which allows events to be injected into SpiNNaker and
    converted in to multicast packets.

    :param str label: The label of this vertex
    :param vertex_slice:
        The slice served via this multicast source
    :type vertex_slice: ~pacman.model.graphs.common.Slice or None
    :param app_vertex:
        The associated application vertex
    :type app_vertex: ReverseIpTagMultiCastSource or None
    :param int n_keys: The number of keys to be sent via this multicast source
        (can't be `None` if vertex_slice is also `None`)
    :param str board_address:
        The IP address of the board on which to place this vertex if receiving
        data, either buffered or live (by default, any board is chosen)
    :param int receive_port:
        The port on the board that will listen for incoming event packets
        (default is to disable this feature; set a value to enable it, or set
        the `reserve_reverse_ip_tag parameter` to True if a random port is to
        be used)
    :param int receive_sdp_port:
        The SDP port to listen on for incoming event packets (defaults to 1)
    :param int receive_tag:
        The IP tag to use for receiving live events (uses any by default)
    :param float receive_rate:
    :param int virtual_key:
        The base multicast key to send received events with (assigned
        automatically by default)
    :param int prefix:
        The prefix to "or" with generated multicast keys (default is no prefix)
    :param ~spinnman.messages.eieio.EIEIOPrefix prefix_type:
        Whether the prefix should apply to the upper or lower half of the
        multicast keys (default is upper half)
    :param bool check_keys:
        True if the keys of received events should be verified before sending
        (default False)
    :param ~numpy.ndarray send_buffer_times:
        An array of arrays of time steps at which keys should be sent (one
        array for each key, default disabled)
    :param str send_buffer_partition_id:
        The ID of the partition containing the edges down which the events are
        to be sent
    :param bool reserve_reverse_ip_tag:
        True if the source should set up a tag through which it can receive
        packets; if port is set to `None` this can be used to enable the
        reception of packets on a randomly assigned port, which can be read
        from the database
    :param str injection_partition:
        If not `None`, will enable injection and specify the partition to send
        injected keys with
    """

    class _REGIONS(IntEnum):
        SYSTEM = 0
        CONFIGURATION = 1
        RECORDING = 2
        SEND_BUFFER = 3
        PROVENANCE_REGION = 4

    class _PROVENANCE_ITEMS(IntEnum):
        N_RECEIVED_PACKETS = 0
        N_SENT_PACKETS = 1
        INCORRECT_KEYS = 2
        INCORRECT_PACKETS = 3
        LATE_PACKETS = 4

    # 13 ints (1. has prefix, 2. prefix, 3. prefix type, 4. check key flag,
    #          5. has key, 6. key, 7. mask, 8. buffer space,
    #          9. send buffer flag before notify, 10. tag,
    #          11. tag destination (y, x), 12. receive SDP port,
    #          13. timer offset)
    _CONFIGURATION_REGION_SIZE = 13 * BYTES_PER_WORD

    # Counts to do timer offsets
    _n_vertices = 0
    _n_data_specs = 0

    def __init__(
            self, label,
            vertex_slice=None,
            app_vertex=None,
            n_keys=None,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value,
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,

            # Extra flag for receiving packets without a port
            reserve_reverse_ip_tag=False,

            # Partition to send injection keys with
            injection_partition_id=None):
        # pylint: disable=too-many-arguments
        if vertex_slice is None:
            if n_keys is not None:
                vertex_slice = Slice(0, n_keys - 1)
            else:
                raise KeyError("Either provide a vertex_slice or n_keys")

        if (send_buffer_partition_id is not None and
                injection_partition_id is not None):
            raise ValueError(
                "Can't specify both send_buffer_partition_id and"
                " injection_partition_id")

        super().__init__(label, app_vertex, vertex_slice)

        self._reverse_iptags = []
        self._n_keys = vertex_slice.n_atoms

        # Set up for receiving live packets
        if receive_port is not None or reserve_reverse_ip_tag:
            self._reverse_iptags = [ReverseIPtagResource(
                port=receive_port, sdp_port=receive_sdp_port,
                tag=receive_tag)]
        self._receive_rate = receive_rate
        self._receive_sdp_port = receive_sdp_port

        # Work out if buffers are being sent
        self._send_buffer = None
        self._first_machine_time_step = None
        self._run_until_timesteps = None
        self._send_buffer_partition_id = send_buffer_partition_id
        self._send_buffer_size = 0
        n_buffer_times = 0
        if send_buffer_times is not None:
            for i in send_buffer_times:
                if hasattr(i, "__len__"):
                    n_buffer_times += len(i)
                else:
                    # assuming this must be a single integer
                    n_buffer_times += 1
            if n_buffer_times == 0:
                logger.warning(
                    "Combination of send_buffer_times {} and slice {} results "
                    "in a core with a ReverseIPTagMulticastSourceMachineVertex"
                    " which does not spike", send_buffer_times, vertex_slice)
        if n_buffer_times == 0:
            self._send_buffer_times = None
            self._send_buffers = None
        else:
            self._install_send_buffer(send_buffer_times)

        # Set up for recording (if requested)
        self._is_recording = False

        # set flag for checking if in injection mode
        self._injection_partition_id = injection_partition_id

        # Sort out the keys to be used
        self._virtual_key = virtual_key
        self._mask = None
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        # Work out the prefix details
        if self._prefix is not None:
            if self._prefix_type is None:
                self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            if self._prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
                self._prefix = prefix << 16

        # If the user has specified a virtual key
        if self._virtual_key is not None:
            self._install_virtual_key(vertex_slice.n_atoms)

        self._n_vertices += 1

    @staticmethod
    def _max_send_buffer_keys_per_timestep(send_buffer_times, n_keys):
        """
        :param send_buffer_times: When events will be sent
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            counts = numpy.bincount(
                numpy.concatenate(send_buffer_times).astype("int"))
            if len(counts):
                return max(counts)
            return 0
        if len(send_buffer_times):
            counts = numpy.bincount(send_buffer_times)
            if len(counts):
                return n_keys * max(counts)
            return 0
        return 0

    @classmethod
    def _send_buffer_sdram_per_timestep(cls, send_buffer_times, n_keys):
        """
        Determine the amount of SDRAM required per timestep.

        :param send_buffer_times:
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        # If there is a send buffer, calculate the keys per timestep
        if send_buffer_times is not None:
            return get_n_bytes(cls._max_send_buffer_keys_per_timestep(
                send_buffer_times, n_keys))
        return 0

    @classmethod
    def _recording_sdram_per_timestep(
            cls, is_recording, receive_rate, send_buffer_times, n_keys):
        """
        :param bool is_recording:
        :param float receive_rate:
        :param send_buffer_times:
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        # If not recording, no SDRAM needed per timestep
        if not is_recording:
            return 0

        # If recording send data, the recorded size is the send size
        if send_buffer_times is not None:
            return cls._send_buffer_sdram_per_timestep(
                send_buffer_times, n_keys)

        # Recording live data, use the user provided receive rate
        # https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon/issues/896
        keys_per_timestep = math.ceil(
            receive_rate / FecDataView.get_simulation_time_step_ms() * 1.1)
        header_size = EIEIODataHeader.get_header_size(
            EIEIOType.KEY_32_BIT, is_payload_base=True)
        # Maximum size is one packet per key
        return ((header_size + EIEIOType.KEY_32_BIT.key_bytes) *
                keys_per_timestep)

    def _install_send_buffer(self, send_buffer_times):
        """
        :param ~numpy.ndarray send_buffer_times:
        """
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            # Working with a list of lists so check length
            if len(send_buffer_times) != self._n_keys:
                raise ConfigurationException(
                    f"The array or arrays of times {send_buffer_times} does "
                    f"not have the expected length of {self._n_keys}")

        self._send_buffer = BufferedSendingRegion()
        self._send_buffer_times = send_buffer_times
        self._send_buffers = {
            self._REGIONS.SEND_BUFFER: self._send_buffer
        }

    def _install_virtual_key(self, n_keys):
        """
        :param int n_keys:
        """
        # check that virtual key is valid
        if self._virtual_key < 0:
            raise ConfigurationException("Virtual keys must be positive")

        # Get a mask and maximum number of keys for the number of keys
        # requested
        self._mask = self.calculate_mask(n_keys)

        if self._prefix is not None:
            # Check that the prefix doesn't change the virtual key in the
            # masked area
            masked_key = (self._virtual_key | self._prefix) & self._mask
            if self._virtual_key != masked_key:
                raise ConfigurationException(
                    "The number of keys, virtual key and key prefix settings "
                    "don't work together")
        else:
            # If no prefix was generated, generate one
            self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            self._prefix = self._virtual_key

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._REGIONS.PROVENANCE_REGION

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return 5

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self):
        return self.get_sdram_usage(
            self._send_buffer_times, self._is_recording,
            self._receive_rate, self._n_keys)

    @property
    @overrides(MachineVertex.reverse_iptags)
    def reverse_iptags(self):
        return self._reverse_iptags

    @classmethod
    def get_sdram_usage(
            cls, send_buffer_times, recording_enabled, receive_rate, n_keys):
        """
        :param send_buffer_times: When events will be sent
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param bool recording_enabled: Whether recording is done
        :param float receive_rate: What the expected message receive rate is
        :param int n_keys: How many keys are being sent
        :rtype: ~pacman.model.resources.VariableSDRAM
        """
        static_usage = (
            SYSTEM_BYTES_REQUIREMENT +
            cls._CONFIGURATION_REGION_SIZE +
            get_recording_header_size(1) +
            get_recording_data_constant_size(1) +
            cls.get_provenance_data_size(0))
        per_timestep = (
            cls._send_buffer_sdram_per_timestep(send_buffer_times, n_keys) +
            cls._recording_sdram_per_timestep(
                recording_enabled, receive_rate, send_buffer_times, n_keys))
        static_usage += per_timestep
        return VariableSDRAM(static_usage, per_timestep)

    @staticmethod
    def _n_regions_to_allocate(send_buffering, recording):
        """
        Get the number of regions that will be allocated

        :param bool send_buffering:
        :param bool recording:
        :rtype: int
        """
        if recording and send_buffering:
            return 5
        if recording or send_buffering:
            return 4
        return 3

    @property
    def send_buffer_times(self):
        """
        When events will be sent.

        :rtype:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        """
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        """
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        """
        self._install_send_buffer(send_buffer_times)

    @staticmethod
    def _is_in_range(step, first_step, end_step):
        """
        :param int step: The time step to check
        :param int first_step: The smallest support step
        :param int end_step: The step after the end
        :type n_machine_time_steps: int or None
        """
        return end_step is None or (first_step <= step < end_step)

    def _fill_send_buffer(self):
        """
        Fill the send buffer with keys to send.
        """
        first_machine_time_step = FecDataView.get_first_machine_time_step()
        run_until_timesteps = FecDataView.get_current_run_timesteps()
        if (self._first_machine_time_step == first_machine_time_step and
                self._run_until_timesteps == run_until_timesteps):
            return
        self._first_machine_time_step = first_machine_time_step
        self._run_until_timesteps = run_until_timesteps
        key_to_send = self._virtual_key
        if self._virtual_key is None:
            key_to_send = 0

        if self._send_buffer is not None:
            self._send_buffer.clear()
        if (self._send_buffer_times is not None and
                len(self._send_buffer_times)):
            if hasattr(self._send_buffer_times[0], "__len__"):
                # Works with a list-of-lists
                self._fill_send_buffer_2d(key_to_send)
            else:
                # Work with a single list
                self._fill_send_buffer_1d(key_to_send)

    def _fill_send_buffer_2d(self, key_base):
        """
        Add the keys with different times for each atom.
        Can be overridden to override keys.

        :param int key_base: The base key to use
        """
        first_time_step = FecDataView.get_first_machine_time_step()
        end_time_step = FecDataView.get_current_run_timesteps()
        if first_time_step == end_time_step:
            return
        keys = get_field_based_keys(key_base, self._vertex_slice)
        for atom in range(self._vertex_slice.n_atoms):
            for tick in sorted(self._send_buffer_times[atom]):
                if self._is_in_range(tick, first_time_step, end_time_step):
                    self._send_buffer.add_key(tick, keys[atom])

    def _fill_send_buffer_1d(self, key_base):
        """
        Add the keys from the given vertex slice within the given time
        range into the given send buffer, with the same times for each
        atom.  Can be overridden to override keys.

        :param int key_base: The base key to use
        """
        first_time_step = FecDataView.get_first_machine_time_step()
        end_time_step = FecDataView.get_current_run_timesteps()
        if first_time_step == end_time_step:
            return
        keys = get_field_based_keys(key_base, self._vertex_slice)
        key_list = [keys[atom] for atom in range(self._vertex_slice.n_atoms)]
        for tick in sorted(self._send_buffer_times):
            if self._is_in_range(tick, first_time_step, end_time_step):
                self._send_buffer.add_keys(tick, key_list)

    @staticmethod
    def _generate_prefix(virtual_key, prefix_type):
        """
        :param ~.EIEIOPrefix prefix_type:
        :param int virtual_key:
        :rtype: int
        """
        if prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return virtual_key & 0xFFFF
        return (virtual_key >> 16) & 0xFFFF

    @staticmethod
    def calculate_mask(n_neurons):
        """
        :param int n_neurons:
        :rtype: int
        """
        temp_value = n_neurons.bit_length()
        max_key = 2**temp_value - 1
        mask = 0xFFFFFFFF - max_key
        return mask

    def enable_recording(self, new_state=True):
        """
        Enable recording of the keys sent.

        :param bool new_state:
        """
        self._is_recording = new_state

    def _reserve_regions(self, spec):
        """
        :param ~.DataSpecificationGenerator spec:
        :param int n_machine_time_steps:
        """
        # Reserve system and configuration memory regions:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES, label='SYSTEM')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIGURATION,
            size=self._CONFIGURATION_REGION_SIZE, label='CONFIGURATION')

        # Reserve recording buffer regions if required
        spec.reserve_memory_region(
            region=self._REGIONS.RECORDING,
            size=get_recording_header_size(1),
            label="RECORDING")

        # Reserve send buffer region if required
        if self._send_buffer_times is not None:
            self._send_buffer_size = (
                self._send_buffer_sdram_per_timestep(
                    self._send_buffer_times, self._n_keys) *
                FecDataView.get_max_run_time_steps())
            if self._send_buffer_size:
                spec.reserve_memory_region(
                    region=self._REGIONS.SEND_BUFFER,
                    size=self._send_buffer_size, label="SEND_BUFFER",
                    empty=True)

        self.reserve_provenance_data_region(spec)

    def update_virtual_key(self):
        routing_info = FecDataView.get_routing_infos()
        if self._virtual_key is None:
            rinfo = None
            if self._send_buffer_partition_id is not None:
                rinfo = routing_info.get_routing_info_from_pre_vertex(
                    self, self._send_buffer_partition_id)
            if self._injection_partition_id is not None:
                rinfo = routing_info.get_routing_info_from_pre_vertex(
                    self, self._injection_partition_id)

            # if no edge leaving this vertex, no key needed
            if rinfo is not None:
                self._virtual_key = rinfo.key
                self._mask = rinfo.mask

        if self._virtual_key is not None and self._prefix is None:
            self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            self._prefix = self._virtual_key

    def _write_configuration(self, spec):
        """
        :param ~.DataSpecificationGenerator spec:
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIGURATION)

        # Write apply_prefix and prefix and prefix_type
        if self._prefix is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._prefix)
            spec.write_value(data=self._prefix_type.value)

        # Write check
        if self._check_keys:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # Write if you have a key to transmit write it and the mask,
        # otherwise write flags to fill in space
        if self._virtual_key is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._virtual_key)
            spec.write_value(data=self._mask)

        # Write send buffer data
        if self._send_buffer_times is not None:

            spec.write_value(data=self._send_buffer_size)
            # The following disable the buffer notifications
            spec.write_value(data=self._send_buffer_size + 256)
            spec.write_value(data=0)
            spec.write_value(_ONE_WORD.unpack(_TWO_SHORTS.pack(0, 0))[0])
        else:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)

        # write SDP port to which SDP packets will be received
        spec.write_value(data=self._receive_sdp_port)

        # write timer offset in microseconds
        max_offset = (FecDataView.get_hardware_time_step_us() // (
            _MAX_OFFSET_DENOMINATOR * 2))
        spec.write_value(
            ((int(math.ceil(max_offset / self._n_vertices)) *
              ReverseIPTagMulticastSourceMachineVertex._n_data_specs) +
             int(math.ceil(max_offset))) % _MAX_OFFSET_MODULO)
        ReverseIPTagMulticastSourceMachineVertex._n_data_specs += 1

    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={"routing_info"})
    def generate_data_specification(
            self, spec, placement):  # @UnusedVariable
        self.update_virtual_key()

        # Reserve regions
        self._reserve_regions(spec)

        # Write the system region
        spec.switch_write_focus(self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name()))

        # Write the additional recording information
        spec.switch_write_focus(self._REGIONS.RECORDING)
        recording_size = 0
        if self._is_recording:
            per_timestep = self._recording_sdram_per_timestep(
                self._is_recording, self._receive_rate,
                self._send_buffer_times, self._n_keys)
            recording_size = (
                    per_timestep * FecDataView.get_max_run_time_steps())
        spec.write_array(get_recording_header_array([recording_size]))

        # Write the configuration information
        self._write_configuration(spec)

        # End spec
        spec.end_specification()

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return "reverse_iptag_multicast_source.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.USES_SIMULATION_INTERFACE

    def get_virtual_key(self):
        """
        Updates and returns the virtual key. `None` is give a zero value

        :rtype: int or None
        """
        self.update_virtual_key()
        if self._virtual_key:
            return self._virtual_key
        return 0

    @property
    def mask(self):
        """
        :rtype: int or None
        """
        return self._mask

    @property
    @overrides(AbstractSupportsDatabaseInjection.is_in_injection_mode)
    def is_in_injection_mode(self):
        return self._injection_partition_id is not None

    @property
    @overrides(AbstractSupportsDatabaseInjection.injection_partition_id)
    def injection_partition_id(self):
        return self._injection_partition_id

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        if not self._is_recording:
            return []
        return [0]

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, placement):
        return locate_memory_region_for_placement(
            placement, self._REGIONS.RECORDING)

    @property
    def send_buffers(self):
        """
        :rtype: dict(int,BufferedSendingRegion)
        """
        self._fill_send_buffer()
        return self._send_buffers

    @overrides(SendsBuffersFromHostPreBufferedImpl.get_regions)
    def get_regions(self):
        # Avoid update_buffer as not needed and called during reset
        return self._send_buffers.keys()

    @overrides(SendsBuffersFromHostPreBufferedImpl.rewind)
    def rewind(self, region):
        # reset theses so fill send buffer will run when send_buffers called
        self._first_machine_time_step = None
        self._run_until_timesteps = None
        # Avoid update_buffer as not needed and called during reset
        self._send_buffers[region].rewind()

    @overrides(SendsBuffersFromHostPreBufferedImpl.buffering_input)
    def buffering_input(self):
        return self._send_buffers is not None

    @send_buffers.setter
    def send_buffers(self, value):
        self._send_buffers = value

    def get_region_buffer_size(self, region):
        """
        :param int region: Region ID
        :return: Size of buffer, in bytes.
        :rtype: int
        """
        if region == self._REGIONS.SEND_BUFFER:
            return self._send_buffer_size
        return 0

    @overrides(
        ProvidesProvenanceDataFromMachineImpl.parse_extra_provenance_items)
    def parse_extra_provenance_items(self, label, x, y, p, provenance_data):
        n_rcv, n_snt, bad_key, bad_pkt, late = provenance_data

        with ProvenanceWriter() as db:
            db.insert_core(x, y, p, "Received_sdp_packets", n_rcv)
            if n_rcv == 0 and self._send_buffer_times is None:
                db.insert_report(
                    f"No SDP packets were received by {label}. "
                    f"If you expected packets to be injected, "
                    f"this could indicate an error")

            db.insert_core(
                x, y, p, "Send_multicast_packets", n_snt)
            if n_snt == 0:
                db.insert_report(
                    f"No multicast packets were sent by {label}. "
                    f"If you expected packets to be sent "
                    f"this could indicate an error")
            db.insert_core(
                x, y, p, "Incorrect_keys", bad_key)
            if bad_key > 0:
                db.insert_report(
                    f"Keys were received by {label} that did not match the "
                    f"key {self._virtual_key} and mask {self._mask}")

            db.insert_core(x, y, p, "Incorrect_packets", bad_pkt)
            if bad_pkt > 0:
                db.insert_report(
                    f"SDP Packets were received by {label} "
                    f"that were not correct")

            db.insert_core(x, y, p, "Late_packets", late)
            if late > 0:
                db.insert_report(
                    f"SDP Packets were received by {label} that were too "
                    f"late to be transmitted in the simulation")

    def __repr__(self):
        return self._label



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, IntEnum
import logging
import struct
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_machine import CoreSubsets, Router
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import ConstantSDRAM
from spinn_utilities.config_holder import get_config_bool
from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.utility_objs.\
    extra_monitor_scp_processes import (
        ReadStatusProcess, ResetCountersProcess, SetPacketTypesProcess,
        SetRouterTimeoutProcess, ClearQueueProcess,
        LoadApplicationMCRoutesProcess, LoadSystemMCRoutesProcess)
from spinn_front_end_common.utilities.constants import (
    SARK_PER_MALLOC_SDRAM_USAGE, DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES,
    BYTES_PER_WORD, BYTES_PER_KB)
from spinn_front_end_common.utilities.helpful_functions import (
    convert_vertices_to_core_subset, get_region_base_address_offset)
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_state_from_failure)
from .data_speed_up_packet_gatherer_machine_vertex import (
    DataSpeedUpPacketGatherMachineVertex as
    Gatherer)
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesProvenanceDataFromMachine, ProvenanceWriter)

log = FormatAdapter(logging.getLogger(__name__))

_CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES = 5 * BYTES_PER_WORD
#: 1.new seq key, 2.first data key, 3. transaction id key 4.end flag key,
# 5.base key
_CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES = 5 * BYTES_PER_WORD
_CONFIG_MAX_EXTRA_SEQ_NUM_SIZE_IN_BYTES = 460 * BYTES_PER_KB
_CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES = 3 * BYTES_PER_WORD
_MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING = ((49 * 3) + 1) * BYTES_PER_WORD
_BIT_SHIFT_TO_MOVE_APP_ID = 24

_ONE_WORD = struct.Struct("<I")
# typedef struct extra_monitor_provenance_t {
#     uint n_sdp_packets;
#     uint n_in_streams;
#     uint n_out_streams;
#     uint n_router_changes;
# } extra_monitor_provenance_t;
_PROVENANCE_FORMAT = struct.Struct("<IIII")

# cap for stopping wrap arounds
TRANSACTION_ID_CAP = 0xFFFFFFFF

# SDRAM requirement for containing router table entries
# 16 bytes per entry:
# 4 for a key, 4 for mask,
# 4 for word alignment for 18 cores and 6 links
# (24 bits, for word aligning)
_SDRAM_FOR_ROUTER_TABLE_ENTRIES = 1024 * 4 * BYTES_PER_WORD


class _DSG_REGIONS(IntEnum):
    REINJECT_CONFIG = 0
    DATA_OUT_CONFIG = 1
    DATA_IN_CONFIG = 2
    PROVENANCE_AREA = 3


class _KEY_OFFSETS(Enum):
    ADDRESS_KEY_OFFSET = 0
    DATA_KEY_OFFSET = 1
    BOUNDARY_KEY_OFFSET = 2


class ExtraMonitorSupportMachineVertex(
        MachineVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification,
        AbstractProvidesProvenanceDataFromMachine):
    """
    Machine vertex for talking to extra monitor cores.
    Supports reinjection control and the faster data transfer protocols.

    Usually deployed once per chip.

    .. note::
        This is an unusual machine vertex, in that it has no associated
        application vertex.
    """

    __slots__ = (
        # if we reinject multicast packets
        "_reinject_multicast",
        # if we reinject point to point packets
        "_reinject_point_to_point",
        # if we reinject nearest neighbour packets
        "_reinject_nearest_neighbour",
        # if we reinject fixed route packets
        "_reinject_fixed_route",
        # placement holder for ease of access
        "_placement",
        # app id, used for reporting failures on system core RTE
        "_app_id",
        # the local transaction id
        "_transaction_id",
        # provenance region address
        "_prov_region"
    )

    def __init__(
            self, reinject_point_to_point=False,
            reinject_nearest_neighbour=False, reinject_fixed_route=False):
        """
        :param bool reinject_point_to_point:
            if we reinject point-to-point packets
        :param bool reinject_nearest_neighbour:
            if we reinject nearest-neighbour packets
        :param bool reinject_fixed_route: if we reinject fixed route packets
        """
        # pylint: disable=too-many-arguments
        super().__init__(
            label="SYSTEM:ExtraMonitor", app_vertex=None)

        self._reinject_multicast = get_config_bool(
            "Machine", "enable_reinjection")
        self._reinject_point_to_point = reinject_point_to_point
        self._reinject_nearest_neighbour = reinject_nearest_neighbour
        self._reinject_fixed_route = reinject_fixed_route
        # placement holder for ease of access
        self._placement = None
        self._app_id = None
        self._transaction_id = 0
        self._prov_region = None

    @property
    def reinject_multicast(self):
        """
        :rtype: bool
        """
        return self._reinject_multicast

    @property
    def transaction_id(self):
        return self._transaction_id

    def update_transaction_id(self):
        self._transaction_id = (self._transaction_id + 1) & TRANSACTION_ID_CAP

    def update_transaction_id_from_machine(self):
        """
        Looks up from the machine what the current transaction id is
        and updates the extra monitor.
        """
        self._transaction_id = FecDataView.get_transceiver().read_user_1(
            self._placement.x, self._placement.y, self._placement.p)

    @property
    def reinject_point_to_point(self):
        """
        :rtype: bool
        """
        return self._reinject_point_to_point

    @property
    def reinject_nearest_neighbour(self):
        """
        :rtype: bool
        """
        return self._reinject_nearest_neighbour

    @property
    def reinject_fixed_route(self):
        """
        :rtype: bool
        """
        return self._reinject_fixed_route

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self):
        return ConstantSDRAM(
            _CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES +
            _CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES +
            _CONFIG_MAX_EXTRA_SEQ_NUM_SIZE_IN_BYTES +
            # Data spec size
            DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES +
            # One malloc for extra sequence numbers
            SARK_PER_MALLOC_SDRAM_USAGE +
            _MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING +
            _SDRAM_FOR_ROUTER_TABLE_ENTRIES +
            _CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES)

    @property
    def placement(self):
        """
        :rtype: ~pacman.model.placements.Placement
        """
        return self._placement

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return self.static_get_binary_start_type()

    @staticmethod
    def static_get_binary_start_type():
        """
        The type of the binary implementing this vertex.

        :rtype: ExecutableType
        """
        return ExecutableType.SYSTEM

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return self.static_get_binary_file_name()

    @staticmethod
    def static_get_binary_file_name():
        """
        The name of the binary implementing this vertex.

        :rtype: str
        """
        return "extra_monitor_support.aplx"

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec, placement):
        # storing for future usage
        self._placement = placement
        self._app_id = FecDataView.get_app_id()
        # write reinjection config
        self._generate_reinjection_config(spec, placement)
        # write data speed up out config
        self._generate_data_speed_up_out_config(spec)
        # write data speed up in config
        self._generate_data_speed_up_in_config(
            spec, FecDataView().get_chip_at(placement.x, placement.y))
        self._generate_provenance_area(spec)
        spec.end_specification()

    def _generate_data_speed_up_out_config(self, spec):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.DATA_OUT_CONFIG,
            size=_CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES,
            label="data speed-up out config region")
        spec.switch_write_focus(_DSG_REGIONS.DATA_OUT_CONFIG)
        spec.write_value(Gatherer.BASE_KEY)
        spec.write_value(Gatherer.NEW_SEQ_KEY)
        spec.write_value(Gatherer.FIRST_DATA_KEY)
        spec.write_value(Gatherer.TRANSACTION_ID_KEY)
        spec.write_value(Gatherer.END_FLAG_KEY)

    def _generate_reinjection_config(self, spec, placement):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        :param ~.Placement placement:
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.REINJECT_CONFIG,
            size=_CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES,
            label="re-injection config region")

        spec.switch_write_focus(_DSG_REGIONS.REINJECT_CONFIG)
        for value in [
                self._reinject_multicast, self._reinject_point_to_point,
                self._reinject_fixed_route,
                self._reinject_nearest_neighbour]:
            # Note that this is inverted! Why... I dunno!
            spec.write_value(int(not value))

        # add the reinjection mc interface
        router_timeout_keys = \
            FecDataView.get_system_multicast_router_timeout_keys()
        chip = FecDataView().get_chip_at(placement.x, placement.y)
        # pylint: disable=unsubscriptable-object
        reinjector_base_mc_key = (
            router_timeout_keys[
                (chip.nearest_ethernet_x, chip.nearest_ethernet_y)])
        spec.write_value(reinjector_base_mc_key)

    def _generate_data_speed_up_in_config(self, spec, chip):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        :param ~.Chip chip: the chip where this monitor will run
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.DATA_IN_CONFIG,
            size=(_MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING +
                  _CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES),
            label="data speed-up in config region")
        spec.switch_write_focus(_DSG_REGIONS.DATA_IN_CONFIG)

        # write address key and data key
        mc_data_chips_to_keys = \
            FecDataView.get_data_in_multicast_key_to_chip_map()
        # pylint: disable=unsubscriptable-object
        base_key = mc_data_chips_to_keys[chip.x, chip.y]
        spec.write_value(base_key + _KEY_OFFSETS.ADDRESS_KEY_OFFSET.value)
        spec.write_value(base_key + _KEY_OFFSETS.DATA_KEY_OFFSET.value)
        spec.write_value(base_key + _KEY_OFFSETS.BOUNDARY_KEY_OFFSET.value)

        # write table entries
        data_in_routing_tables = \
            FecDataView.get_data_in_multicast_routing_tables()
        table = data_in_routing_tables.get_routing_table_for_chip(
            chip.x, chip.y)
        spec.write_value(table.number_of_entries)
        for entry in table.multicast_routing_entries:
            spec.write_value(entry.routing_entry_key)
            spec.write_value(entry.mask)
            spec.write_value(self.__encode_route(entry))

    def __encode_route(self, entry):
        """
        :param ~spinn_machine.MulticastRoutingEntry entry:
        :rtype: int
        """
        route = self._app_id << _BIT_SHIFT_TO_MOVE_APP_ID
        route |= Router.convert_routing_table_entry_to_spinnaker_route(entry)
        return route

    def _generate_provenance_area(self, spec):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.PROVENANCE_AREA, size=_PROVENANCE_FORMAT.size,
            label="provenance collection region", empty=True)

    def __get_provenance_region_address(self, txrx, place):
        """
        :param ~spinnman.transceiver.Transceiver txrx:
        :param ~pacman.model.placements.Placement place:
        :rtype: int
        """
        if self._prov_region is None:
            region_table_addr = txrx.get_cpu_information_from_core(
                place.x, place.y, place.p).user[0]
            region_entry_addr = get_region_base_address_offset(
                region_table_addr, _DSG_REGIONS.PROVENANCE_AREA)
            self._prov_region, = _ONE_WORD.unpack(txrx.read_memory(
                place.x, place.y, region_entry_addr, BYTES_PER_WORD))
        return self._prov_region

    @overrides(AbstractProvidesProvenanceDataFromMachine.
               get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, placement):
        # No standard provenance region, so no standard provenance data
        # But we do have our own.
        transceiver = FecDataView.get_transceiver()
        provenance_address = self.__get_provenance_region_address(
            transceiver, placement)
        data = transceiver.read_memory(
            placement.x, placement.y, provenance_address,
            _PROVENANCE_FORMAT.size)
        (n_sdp_packets, n_in_streams, n_out_streams, n_router_changes) = \
            _PROVENANCE_FORMAT.unpack_from(data)
        with ProvenanceWriter() as db:
            db.insert_monitor(
                placement.x, placement.y,
                "Number_of_Router_Configuration_Changes", n_router_changes)
            db.insert_monitor(
                placement.x, placement.y,
                "Number_of_Relevant_SDP_Messages", n_sdp_packets)
            db.insert_monitor(
                placement.x, placement.y,
                "Number_of_Input_Streamlets", n_in_streams)
            db.insert_monitor(
                placement.x, placement.y,
                "Number_of_Output_Streamlets", n_out_streams)

    def set_router_wait1_timeout(
            self, timeout, extra_monitor_cores_to_set):
        """
        Supports setting of the router time outs for a set of chips via their
        extra monitor cores. This sets the timeout for the time between when a
        packet arrives and when it starts to be emergency routed. (Actual
        emergency routing is disabled by default.)

        :param tuple(int,int) timeout:
            The mantissa and exponent of the timeout value, each between
            0 and 15
        :param extra_monitor_cores_to_set:
            which monitors control the routers to set the timeout of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set)
        process = SetRouterTimeoutProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.set_wait1_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def set_router_wait2_timeout(
            self, timeout, extra_monitor_cores_to_set):
        """
        Supports setting of the router time outs for a set of chips via their
        extra monitor cores. This sets the timeout for the time between when a
        packet starts to be emergency routed and when it is dropped. (Actual
        emergency routing is disabled by default.)

        :param tuple(int,int) timeout:
            The mantissa and exponent of the timeout value, each between
            0 and 15
        :param extra_monitor_cores_to_set:
            which monitors control the routers to set the timeout of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set)
        process = SetRouterTimeoutProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.set_wait2_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def reset_reinjection_counters(self, extra_monitor_cores_to_set):
        """
        Resets the counters for reinjection.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :param extra_monitor_cores_to_set:
            which monitors control the routers to reset the counters of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set)
        process = ResetCountersProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def clear_reinjection_queue(self, extra_monitor_cores_to_set):
        """
        Clears the queues for reinjection.

        :param extra_monitor_cores_to_set:
            Which extra monitors need to clear their queues.
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set)
        process = ClearQueueProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def get_reinjection_status(self):
        """
        Get the reinjection status from this extra monitor vertex.

        :return: the reinjection status for this vertex
        :rtype: ReInjectionStatus
        """
        placement = FecDataView.get_placement_of_vertex(self)
        process = ReadStatusProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            return process.get_reinjection_status(
                placement.x, placement.y, placement.p)
        except:  # noqa: E722
            emergency_recover_state_from_failure(self, placement)
            raise

    def get_reinjection_status_for_vertices(self):
        """
        Get the reinjection status from a set of extra monitor cores.

        :rtype: dict(tuple(int,int), ReInjectionStatus)
        """
        core_subsets = convert_vertices_to_core_subset(
            FecDataView.iterate_monitors())
        process = ReadStatusProcess(
            FecDataView.get_scamp_connection_selector())
        return process.get_reinjection_status_for_core_subsets(core_subsets)

    def set_reinjection_packets(
            self, point_to_point=None, multicast=None, nearest_neighbour=None,
            fixed_route=None):
        """
        :param point_to_point:
            If point to point should be set, or `None` if left as before
        :type point_to_point: bool or None
        :param multicast:
            If multicast should be set, or `None` if left as before
        :type multicast: bool or None
        :param nearest_neighbour:
            If nearest neighbour should be set, or `None` if left as before
        :type nearest_neighbour: bool or None
        :param fixed_route:
            If fixed route should be set, or `None` if left as before.
        :type fixed_route: bool or None
        """
        # pylint: disable=too-many-arguments
        if multicast is not None:
            self._reinject_multicast = multicast
        if point_to_point is not None:
            self._reinject_point_to_point = point_to_point
        if nearest_neighbour is not None:
            self._reinject_nearest_neighbour = nearest_neighbour
        if fixed_route is not None:
            self._reinject_fixed_route = fixed_route

        core_subsets = convert_vertices_to_core_subset(
            FecDataView.iterate_monitors())
        process = SetPacketTypesProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.set_packet_types(
                core_subsets, self._reinject_point_to_point,
                self._reinject_multicast, self._reinject_nearest_neighbour,
                self._reinject_fixed_route)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def load_system_mc_routes(self):
        """
        Get the extra monitor cores to load up the system-based
        multicast routes (used by the Data In protocol).

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        """
        core_subsets = self._convert_vertices_to_core_subset()
        process = LoadSystemMCRoutesProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            return process.load_system_mc_routes(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def load_application_mc_routes(self):
        """
        Get the extra monitor cores to load up the application-based
        multicast routes (used by the Data In protocol).
        """
        core_subsets = self._convert_vertices_to_core_subset()
        process = LoadApplicationMCRoutesProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            return process.load_application_mc_routes(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    @staticmethod
    def _convert_vertices_to_core_subset():
        """
        Convert vertices into the subset of cores where they've been placed.

        :return: where the vertices have been placed
        :rtype: ~.CoreSubsets
        """
        core_subsets = CoreSubsets()
        for vertex in FecDataView.iterate_monitors():
            placement = FecDataView.get_placement_of_vertex(vertex)
            core_subsets.add_processor(placement.x, placement.y, placement.p)
        return core_subsets



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from spinn_utilities.log import FormatAdapter
from spinnman.constants import POWER_CYCLE_WAIT_TIME_IN_SECONDS
from spinnman.transceiver import create_transceiver_from_hostname
from spinnman.model import BMPConnectionData
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import ConfigurationException
import time
import logging

logger = FormatAdapter(logging.getLogger(__name__))

POWER_CYCLE_WARNING = (
    "When power-cycling a board, it is recommended that you wait for 30 "
    "seconds before attempting a reboot. Therefore, the tools will now "
    "wait for 30 seconds. If you wish to avoid this wait, please set "
    "reset_machine_on_startup = False in the [Machine] section of the "
    "relevant configuration (cfg) file.")

POWER_CYCLE_FAILURE_WARNING = (
    "The end user requested the power-cycling of the board. But the "
    "tools did not have the required BMP connection to facilitate a "
    "power-cycling, and therefore will not do so. please set the "
    "bmp_names accordingly in the [Machine] section of the relevant "
    "configuration (cfg) file. Or use a machine assess process which "
    "provides the BMP data (such as a spalloc system) or finally set "
    "reset_machine_on_startup = False in the [Machine] section of the "
    "relevant configuration (cfg) file to avoid this warning in future.")


def machine_generator(
        bmp_details, board_version, auto_detect_bmp,
        scamp_connection_data, reset_machine_on_start_up):
    """
    Makes a transceiver and a machine object.

    :param str bmp_details: the details of the BMP connections
    :param int board_version:
        the version of the boards being used within the machine
        (1, 2, 3, 4 or 5)
    :param bool auto_detect_bmp:
        Whether the BMP should be automatically determined
    :param scamp_connection_data:
        Job.connection dict, a String SC&MP connection data or `None`
    :type scamp_connection_data:
        dict((int,int), str) or None
    :param bool reset_machine_on_start_up:
    :param MachineAllocationController allocation_controller:
        The allocation controller; in some cases, we delegate the creation of
        the transceiver to it.
    :return: Transceiver, and description of machine it is connected to
    :rtype: tuple(~spinn_machine.Machine,
        ~spinnman.transceiver.Transceiver)
    """
    # pylint: disable=too-many-arguments
    if FecDataView.has_allocation_controller():
        # If there is an allocation controller and it wants to make a
        # transceiver for us, we let it do so; transceivers obtained that way
        # are already fully configured
        txrx = FecDataView.get_allocation_controller().create_transceiver()
        if txrx:
            return txrx.get_machine_details(), txrx

    txrx = create_transceiver_from_hostname(
        hostname=FecDataView.get_ipaddress(),
        bmp_connection_data=_parse_bmp_details(bmp_details),
        version=board_version,
        auto_detect_bmp=auto_detect_bmp)

    if reset_machine_on_start_up:
        success = txrx.power_off_machine()
        if success:
            logger.warning(POWER_CYCLE_WARNING)
            time.sleep(POWER_CYCLE_WAIT_TIME_IN_SECONDS)
            logger.warning("Power cycle wait complete")
        else:
            logger.warning(POWER_CYCLE_FAILURE_WARNING)

    # do auto boot if possible
    if board_version is None:
        raise ConfigurationException(
            "Please set a machine version number in the "
            "corresponding configuration (cfg) file")
    txrx.ensure_board_is_ready()
    if scamp_connection_data:
        txrx.add_scamp_connections(scamp_connection_data)
    else:
        txrx.discover_scamp_connections()
    return txrx.get_machine_details(), txrx


def _parse_bmp_cabinet_and_frame(bmp_cabinet_and_frame):
    """
    :param str bmp_cabinet_and_frame:
    :rtype: tuple(int or str, int or str, str, str or None)
    """
    split_string = bmp_cabinet_and_frame.split(";", 2)
    if len(split_string) == 1:
        host = split_string[0].split(",")
        if len(host) == 1:
            return 0, 0, split_string[0], None
        return 0, 0, host[0], host[1]
    if len(split_string) == 2:
        host = split_string[1].split(",")
        if len(host) == 1:
            return 0, split_string[0], host[0], None
        return 0, split_string[0], host[0], host[1]
    host = split_string[2].split(",")
    if len(host) == 1:
        return split_string[0], split_string[1], host[0], None
    return split_string[0], split_string[1], host[0], host[1]


def _parse_bmp_boards(bmp_boards):
    """
    :param str bmp_boards:
    :rtype: list(int)
    """
    # If the string is a range of boards, get the range
    range_match = re.match(r"(\d+)-(\d+)", bmp_boards)
    if range_match is not None:
        return list(range(int(range_match.group(1)),
                          int(range_match.group(2)) + 1))

    # Otherwise, assume a list of boards
    return [int(board) for board in bmp_boards.split(",")]


def _parse_bmp_connection(bmp_detail):
    """
    Parses one item of BMP connection data. Maximal format:
    `cabinet;frame;host,port/boards`

    All parts except host can be omitted. Boards can be a
    hyphen-separated range or a comma-separated list.

    :param str bmp_detail:
    :rtype: ~.BMPConnectionData
    """
    pieces = bmp_detail.split("/")
    (cabinet, frame, hostname, port_num) = \
        _parse_bmp_cabinet_and_frame(pieces[0])
    # if there is no split, then assume its one board, located at 0
    boards = [0] if len(pieces) == 1 else _parse_bmp_boards(pieces[1])
    port_num = None if port_num is None else int(port_num)
    return BMPConnectionData(cabinet, frame, hostname, boards, port_num)


def _parse_bmp_details(bmp_string):
    """
    Take a BMP line (a colon-separated list) and split it into the
    BMP connection data.

    :param str bmp_string: the BMP string to be converted
    :return: the BMP connection data
    :rtype: list(~.BMPConnectionData) or None
    """
    if bmp_string is None or bmp_string == "None":
        return None
    return [_parse_bmp_connection(bmp_connection)
            for bmp_connection in bmp_string.split(":")]



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_front_end_common.data.fec_data_view import FecDataView
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utility_models import CommandSender
from pacman.model.graphs.application import ApplicationVirtualVertex
from pacman.model.placements import Placement
from pacman.model.partitioner_splitters import SplitterOneAppOneMachine
from spinn_utilities.progress_bar import ProgressBar


def add_command_senders(system_placements):
    """
    Add command senders
    """
    return CommandSenderAdder(system_placements).add_command_senders()


class CommandSenderAdder(object):
    __slots__ = [
        "__command_sender_for_chip",
        "__general_command_sender",
        "__system_placements"
    ]

    def __init__(self, system_placements):
        self.__system_placements = system_placements

        # Keep track of command senders by which chip they are on
        self.__command_sender_for_chip = dict()
        self.__general_command_sender = None

    def add_command_senders(self):
        progress = ProgressBar(FecDataView.get_n_vertices(), "Adding commands")
        for vertex in progress.over(FecDataView.iterate_vertices()):
            if isinstance(vertex, AbstractSendMeMulticastCommandsVertex):
                machine = FecDataView.get_machine()
                link_data = None

                # See if we need a specific placement for a device
                if isinstance(vertex, ApplicationVirtualVertex):
                    link_data = vertex.get_outgoing_link_data(machine)

                command_sender = self.__get_command_sender(link_data)

                # allow the command sender to create key to partition map
                command_sender.add_commands(
                    vertex.start_resume_commands,
                    vertex.pause_stop_commands,
                    vertex.timed_commands, vertex)

        all_command_senders = list(self.__command_sender_for_chip.values())
        if self.__general_command_sender is not None:
            all_command_senders.append(self.__general_command_sender)

        return all_command_senders

    def __cores(self, x, y):
        return [p.processor_id
                for p in FecDataView.get_chip_at(x, y).processors
                if not p.is_monitor]

    def __get_command_sender(self, link_data):
        if link_data is None:
            if self.__general_command_sender is None:
                self.__general_command_sender = self.__new_command_sender(
                    "General command sender")
            return self.__general_command_sender

        x = link_data.connected_chip_x
        y = link_data.connected_chip_y

        command_sender = self.__command_sender_for_chip.get((x, y))
        if command_sender is None:
            command_sender = self.__new_command_sender(
                f"Command Sender on {x}, {y}")
            self.__command_sender_for_chip[(x, y)] = command_sender
            cores = self.__cores(x, y)
            p = cores[self.__system_placements.n_placements_on_chip(x, y)]
            self.__system_placements.add_placement(
                Placement(command_sender.machine_vertex, x, y, p))
        return command_sender

    def __new_command_sender(self, label):
        command_sender = CommandSender(label)
        command_sender.splitter = SplitterOneAppOneMachine()
        return command_sender



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification)
from spinn_front_end_common.data import FecDataView


class MachineDataSpecableVertex(
        AbstractGeneratesDataSpecification, allow_derivation=True):
    """
    Support for a vertex that simplifies generating a data specification.
    """
    __slots__ = ()

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec, placement):
        # pylint: disable=too-many-arguments, arguments-differ
        tags = FecDataView.get_tags()
        iptags = tags.get_ip_tags_for_vertex(placement.vertex)
        reverse_iptags = tags.get_reverse_ip_tags_for_vertex(placement.vertex)
        self.generate_machine_data_specification(
            spec, placement, iptags, reverse_iptags)

    @abstractmethod
    def generate_machine_data_specification(
            self, spec, placement, iptags, reverse_iptags):
        """
        :param ~data_specification.DataSpecificationGenerator spec:
            The data specification to write into.
        :param ~pacman.model.placements.Placement placement:
            Where this node is on the SpiNNaker machine.
        :param iptags: The (forward) IP tags for the vertex, if any
        :type iptags: iterable(~spinn_machine.tags.IPTag) or None
        :param reverse_iptags: The reverse IP tags for the vertex, if any
        :type reverse_iptags:
            iterable(~spinn_machine.tags.ReverseIPTag) or None
        """
        # pylint: disable=too-many-arguments



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.placements import Placement
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utility_models import (
    DataSpeedUpPacketGatherMachineVertex, ExtraMonitorSupportMachineVertex)


def insert_extra_monitor_vertices_to_graphs(placements):
    """
    Inserts the extra monitor vertices into the graph that correspond to
    the extra monitor cores required.

    :return: vertex to Ethernet connection map,
        list of extra_monitor_vertices,
        vertex_to_chip_map
    :rtype: tuple(
        dict(tuple(int,int),DataSpeedUpPacketGatherMachineVertex),
        list(ExtraMonitorSupportMachineVertex),
        dict(tuple(int,int),ExtraMonitorSupportMachineVertex))
    """
    # pylint: disable=too-many-arguments, attribute-defined-outside-init
    chip_to_gatherer_map = dict()
    vertex_to_chip_map = dict()
    machine = FecDataView.get_machine()
    ethernet_chips = list(machine.ethernet_connected_chips)
    progress = ProgressBar(
        len(ethernet_chips), "Inserting extra monitors into graphs")

    for eth in progress.over(machine.ethernet_connected_chips):
        gatherer = DataSpeedUpPacketGatherMachineVertex(
            x=eth.x, y=eth.y, ip_address=eth.ip_address)
        chip_to_gatherer_map[eth.x, eth.y] = gatherer
        cores = __cores(machine, eth.x, eth.y)
        p = cores[placements.n_placements_on_chip(eth.x, eth.y)]
        placements.add_placement(Placement(gatherer, eth.x, eth.y, p))
        for x, y in machine.get_existing_xys_by_ethernet(eth.x, eth.y):
            monitor = ExtraMonitorSupportMachineVertex()
            vertex_to_chip_map[x, y] = monitor
            cores = __cores(machine, x, y)
            p = cores[placements.n_placements_on_chip(x, y)]
            placements.add_placement(Placement(monitor, x, y, p))
    return chip_to_gatherer_map, vertex_to_chip_map


def __cores(machine, x, y):
    return [p.processor_id for p in machine.get_chip_at(x, y).processors
            if not p.is_monitor]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
from spinn_utilities.log_store import LogStore
from spinn_utilities.overrides import overrides
from .global_provenance import GlobalProvenance


class LogStoreDB(LogStore):

    @overrides(LogStore.store_log)
    def store_log(self, level, message, timestamp=None):
        try:
            with GlobalProvenance() as db:
                db.store_log(level, message, timestamp)
        except sqlite3.OperationalError as ex:
            if "database is locked" in ex.args:
                # Ok ignore this one
                # DO NOT log this error here or you will loop forever!
                return
            # all others are bad
            raise

    @overrides(LogStore.retreive_log_messages)
    def retreive_log_messages(self, min_level=0):
        with GlobalProvenance() as db:
            return db.retreive_log_messages(min_level)

    @overrides(LogStore.get_location)
    def get_location(self):
        return GlobalProvenance.get_global_provenace_path()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict


class PowerUsed(object):
    """
    Describes the power used by a simulation.
    """

    __slots__ = [
        "__num_chips",
        "__num_cores",
        "__num_fpgas",
        "__num_frames",
        "__exec_time",
        "__mapping_time",
        "__data_gen_time",
        "__loading_time",
        "__saving_time",
        "__chip_energy",
        "__fpga_total_energy",
        "__fpga_exec_energy",
        "__baseline_energy",
        "__packet_energy",
        "__mapping_energy",
        "__data_gen_energy",
        "__loading_energy",
        "__saving_energy",
        "__core_energy",
        "__router_energy"]

    def __init__(self):
        self.__num_chips = 0
        self.__num_cores = 0
        self.__num_fpgas = 0
        self.__num_frames = 0
        self.__exec_time = 0.0
        self.__mapping_time = 0.0
        self.__data_gen_time = 0.0
        self.__loading_time = 0.0
        self.__saving_time = 0.0
        self.__chip_energy = 0.0
        self.__fpga_total_energy = 0.0
        self.__fpga_exec_energy = 0.0
        self.__baseline_energy = 0.0
        self.__packet_energy = 0.0
        self.__mapping_energy = 0.0
        self.__data_gen_energy = 0.0
        self.__loading_energy = 0.0
        self.__saving_energy = 0.0
        self.__core_energy = defaultdict(float)
        self.__router_energy = defaultdict(float)

    @property
    def num_chips(self):
        """
        The total number of chips used.

        :rtype: int
        """
        return self.__num_chips

    @num_chips.setter
    def num_chips(self, value):
        self.__num_chips = int(value)

    @property
    def num_cores(self):
        """
        The total number of cores used, including for SCAMP.

        :rtype: int
        """
        return self.__num_cores

    @num_cores.setter
    def num_cores(self, value):
        self.__num_cores = int(value)

    @property
    def num_fpgas(self):
        """
        The total number of FPGAs used.

        :rtype: int
        """
        return self.__num_fpgas

    @num_fpgas.setter
    def num_fpgas(self, value):
        self.__num_fpgas = int(value)

    @property
    def num_frames(self):
        """
        The total number of frames used.

        :rtype: int
        """
        return self.__num_frames

    @num_frames.setter
    def num_frames(self, value):
        self.__num_frames = int(value)

    @property
    def total_time_secs(self):
        """
        Time taken in total, in seconds.

        :rtype: float
        """
        return self.exec_time_secs + self.loading_time_secs + \
            self.saving_time_secs + self.data_gen_time_secs + \
            self.mapping_time_secs

    @property
    def booted_time_secs(self):
        """
        Time taken when the machine is booted, in seconds.

        :rtype: float
        """
        return self.exec_time_secs + self.loading_time_secs + \
            self.saving_time_secs

    @property
    def exec_time_secs(self):
        """
        Time taken by active simulation running, in seconds.

        :rtype: float
        """
        return self.__exec_time

    @exec_time_secs.setter
    def exec_time_secs(self, value):
        self.__exec_time = float(value)

    @property
    def mapping_time_secs(self):
        """
        Time taken by the mapping phase, in seconds.

        :rtype: float
        """
        return self.__mapping_time

    @mapping_time_secs.setter
    def mapping_time_secs(self, value):
        self.__mapping_time = float(value)

    @property
    def data_gen_time_secs(self):
        """
        Time taken by data generation phase, in seconds.

        :rtype: float
        """
        return self.__data_gen_time

    @data_gen_time_secs.setter
    def data_gen_time_secs(self, value):
        self.__data_gen_time = float(value)

    @property
    def loading_time_secs(self):
        """
        Time taken by data loading, in seconds.

        :rtype: float
        """
        return self.__loading_time

    @loading_time_secs.setter
    def loading_time_secs(self, value):
        self.__loading_time = float(value)

    @property
    def saving_time_secs(self):
        """
        Time taken by data extraction, in seconds.

        :rtype: float
        """
        return self.__saving_time

    @saving_time_secs.setter
    def saving_time_secs(self, value):
        self.__saving_time = float(value)

    @property
    def total_energy_joules(self):
        """
        Total of all energy costs, in Joules.

        :rtype: float
        """
        baseline_energy = self.baseline_joules + self.fpga_total_energy_joules
        idle_energy = self.mapping_joules + self.data_gen_joules
        active_energy = self.chip_energy_joules + self.packet_joules + \
            self.loading_joules + self.saving_joules
        return baseline_energy + idle_energy + active_energy

    @property
    def chip_energy_joules(self):
        """
        Energy used by all SpiNNaker chips during active simulation
        running, in Joules.

        :rtype: float
        """
        return self.__chip_energy

    @chip_energy_joules.setter
    def chip_energy_joules(self, value):
        self.__chip_energy = float(value)

    @property
    def fpga_total_energy_joules(self):
        """
        Energy used by all FPGAs in total, in Joules.

        :rtype: float
        """
        return self.__fpga_total_energy

    @fpga_total_energy_joules.setter
    def fpga_total_energy_joules(self, value):
        self.__fpga_total_energy = float(value)

    @property
    def fpga_exec_energy_joules(self):
        """
        Energy used by all FPGAs during active simulation running, in
        Joules. This is *included* in the total FPGA energy.

        :rtype: float
        """
        return self.__fpga_exec_energy

    @fpga_exec_energy_joules.setter
    def fpga_exec_energy_joules(self, value):
        self.__fpga_exec_energy = float(value)

    @property
    def baseline_joules(self):
        """
        Baseline/idle energy used, in Joules. This is used by things like the
        frames the SpiNNaker boards are held in, the cooling system, etc.

        :rtype: float
        """
        return self.__baseline_energy

    @baseline_joules.setter
    def baseline_joules(self, value):
        self.__baseline_energy = float(value)

    @property
    def packet_joules(self):
        """
        Energy used by packet transmission, in Joules.

        :rtype: float
        """
        return self.__packet_energy

    @packet_joules.setter
    def packet_joules(self, value):
        self.__packet_energy = float(value)

    @property
    def mapping_joules(self):
        """
        Energy used during the mapping phase, in Joules. Assumes that
        the SpiNNaker system has been shut down.

        :rtype: float
        """
        return self.__mapping_energy

    @mapping_joules.setter
    def mapping_joules(self, value):
        self.__mapping_energy = float(value)

    @property
    def data_gen_joules(self):
        """
        Energy used during the data generation phase, in Joules. Assumes
        that the SpiNNaker system has been shut down.

        :rtype: float
        """
        return self.__data_gen_energy

    @data_gen_joules.setter
    def data_gen_joules(self, value):
        self.__data_gen_energy = float(value)

    @property
    def loading_joules(self):
        """
        Energy used during data loading, in Joules.

        :rtype: float
        """
        return self.__loading_energy

    @loading_joules.setter
    def loading_joules(self, value):
        self.__loading_energy = float(value)

    @property
    def saving_joules(self):
        """
        Energy used during data extraction, in Joules.

        :rtype: float
        """
        return self.__saving_energy

    @saving_joules.setter
    def saving_joules(self, value):
        self.__saving_energy = float(value)

    def get_router_active_energy_joules(self, x, y):
        """
        Energy used (above idle baseline) by a particular router, in Joules.

        Unused routers always report 0.0 for this.

        :param int x:
        :param int y:
        :rtype: float
        """
        return self.__router_energy[x, y]

    def add_router_active_energy(self, x, y, joules):
        """
        Adds energy for a particular router.
        It can be called multiple times per router.

        Only intended to be used during construction of this object.

        :param int x:
        :param int y:
        :param float joules: the energy to add for this router, in Joules.
        """
        self.__router_energy[x, y] += float(joules)

    @property
    def active_routers(self):
        """
        Enumeration of the coordinates of the routers that can report
        active energy usage.

        :rtype: iterable(tuple(int, int))
        """
        return self.__router_energy.keys()

    def get_core_active_energy_joules(self, x, y, p):
        """
        Energy used (above idle baseline) by a particular core, in Joules.

        Unused cores always report 0.0 for this.

        :param int x:
        :param int y:
        :param int p:
        :rtype: float
        """
        return self.__core_energy[x, y, p]

    def add_core_active_energy(self, x, y, p, joules):
        """
        Adds energy for a particular core.
        It can be called multiple times per core.

        Only intended to be used during construction of this object.

        :param int x:
        :param int y:
        :param int p:
        :param float joules: the energy to add for this core, in Joules.
        """
        self.__core_energy[x, y, p] += float(joules)

    @property
    def active_cores(self):
        """
        Enumeration of the coordinates of the cores that can report active
        energy usage.

        :rtype: iterable(tuple(int, int, int))
        """
        return self.__core_energy.keys()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from spinn_front_end_common.interface.provenance import ProvenanceWriter

#: The simple properties of PowerUsed object to be reported
_BASIC_PROPERTIES = (
    # Counts
    "num_chips", "num_cores", "num_fpgas", "num_frames",
    # Times (in seconds)
    "total_time_secs", "booted_time_secs", "mapping_time_secs",
    "data_gen_time_secs", "loading_time_secs", "exec_time_secs",
    "saving_time_secs",
    # Energies (in Joules)
    "total_energy_joules", "baseline_joules",
    "fpga_total_energy_joules", "fpga_exec_energy_joules",
    "packet_joules", "mapping_joules", "data_gen_joules",
    "loading_joules", "chip_energy_joules", "saving_joules")
#: The main provenance key we use
_PROV_KEY = "power_provenance"


def energy_provenance_reporter(power_used):
    """
    Converts the power usage information into provenance data.

    :param PowerUsed power_used:
        The computed basic power consumption information
    """
    with ProvenanceWriter() as db:
        for prop in _BASIC_PROPERTIES:
            db.insert_power(
                __prop_name(prop), getattr(power_used, prop))
            for x, y, p in power_used.active_cores:
                db.insert_core(
                    x, y, p, "Energy (Joules)",
                    power_used.get_core_active_energy_joules(x, y, p))
            for x, y in power_used.active_routers:
                db.insert_router(
                    x, y, "Energy (Joules)",
                    power_used.get_router_active_energy_joules(x, y))


def __prop_name(name):
    name = name.capitalize()
    name = re.sub(r"_time_secs$", r" time (seconds)", name)
    return re.sub(r"(_energy)?_joules", r" energy (Joules)", name)


def __router_name(x, y):
    return f"router@{x},{y} energy (Joules)"



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import distutils.dir_util
from setuptools import setup
import os
import sys


if __name__ == '__main__':
    # Repeated installs assume files have not changed
    # https://github.com/pypa/setuptools/issues/3236
    if len(sys.argv) > 0 and sys.argv[1] == 'egg_info':
        # on the first call to setpy.py remove files left by previous install
        this_dir = os.path.dirname(os.path.abspath(__file__))
        build_dir = os.path.join(this_dir, "build")
        if os.path.isdir(build_dir):
            distutils.dir_util.remove_tree(build_dir)
        egg_dir = os.path.join(this_dir, "SpiNNFrontEndCommon.egg-info")
        if os.path.isdir(egg_dir):
            distutils.dir_util.remove_tree(egg_dir)
    setup()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
import logging
from spinn_utilities.log import FormatAdapter
from pacman.exceptions import (PacmanRoutingException)
from pacman.model.routing_tables import (
    MulticastRoutingTables, UnCompressedMulticastRoutingTable)
from spinn_machine import MulticastRoutingEntry
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView

# ADDRESS_KEY, DATA_KEY, BOUNDARY_KEY
N_KEYS_PER_PARTITION_ID = 8
# timeout key, timeout emergency key, clear reinjection queue.
N_KEYS_PER_REINJECTION_PARTITION = 3
KEY_START_VALUE = 8
ROUTING_MASK = 0xFFFFFFF8

logger = FormatAdapter(logging.getLogger(__name__))


def system_multicast_routing_generator():
    """
    Generates routing table entries used by the data-in processes with the
    extra monitor cores.

    :return: routing tables, destination-to-key map,
        board-location-to-timeout-key map
    :rtype: tuple(~pacman.model.routing_tables.MulticastRoutingTables,
        dict(tuple(int,int),int), dict(tuple(int,int),int))
    """
    generator = _SystemMulticastRoutingGenerator()
    # pylint: disable=protected-access
    return generator._run()


class _SystemMulticastRoutingGenerator(object):
    """
    Generates routing table entries used by the data in processes with the
    extra monitor cores.
    """
    __slots__ = ["_key_to_destination_map", "_machine",
                 "_routing_tables", "_time_out_keys_by_board"]

    def __init__(self):
        """

        :param ~pacman.model.placements.Placements placements:
        """
        self._machine = FecDataView.get_machine()
        self._routing_tables = MulticastRoutingTables()
        self._key_to_destination_map = dict()
        self._time_out_keys_by_board = dict()

    def _run(self):
        """
        :return: routing tables, destination-to-key map,
            board-location-to-timeout-key map
        :rtype: tuple(~pacman.model.routing_tables.MulticastRoutingTables,
            dict(tuple(int,int),int), dict(tuple(int,int),int))
        """
        # create progress bar
        progress = ProgressBar(
            self._machine.ethernet_connected_chips,
            "Generating routing tables for data in system processes")

        for ethernet_chip in progress.over(
                self._machine.ethernet_connected_chips):
            tree = self._generate_routing_tree(ethernet_chip)
            if tree is None:
                tree = self._logging_retry(ethernet_chip)
            self._add_routing_entries(ethernet_chip, tree)

        return (self._routing_tables, self._key_to_destination_map,
                self._time_out_keys_by_board)

    def _generate_routing_tree(self, ethernet_chip):
        """
        Generates a map for each chip to over which link it gets its data.

        :param ~spinn_machine.Chip ethernet_chip:
        :return: Map of chip.x, chip.y to (source.x, source.y, source.link)
        :rtype: dict(tuple(int, int), tuple(int, int, int))
        """
        eth_x = ethernet_chip.x
        eth_y = ethernet_chip.y
        tree = dict()

        to_reach = set(
            self._machine.get_existing_xys_by_ethernet(eth_x, eth_y))
        reached = set()
        reached.add((eth_x, eth_y))
        to_reach.remove((eth_x, eth_y))
        found = set()
        found.add((eth_x, eth_y))
        while len(to_reach) > 0:
            just_reached = found
            found = set()
            for x, y in just_reached:
                # Check links starting with the most direct from 0,0
                for link_id in [1, 0, 2, 5, 3, 4]:
                    # Get protential destination
                    destination = self._machine.xy_over_link(x, y, link_id)
                    # If it is useful
                    if destination in to_reach:
                        # check it actually exits
                        if self._machine.is_link_at(x, y, link_id):
                            # Add to tree and record chip reachable
                            tree[destination] = (x, y, link_id)
                            to_reach.remove(destination)
                            found.add(destination)
            if len(found) == 0:
                return None
        return tree

    def _logging_retry(self, ethernet_chip):
        eth_x = ethernet_chip.x
        eth_y = ethernet_chip.y
        tree = dict()
        to_reach = set(
            self._machine.get_existing_xys_by_ethernet(eth_x, eth_y))
        reached = set()
        reached.add((eth_x, eth_y))
        to_reach.remove((eth_x, eth_y))
        found = set()
        found.add((eth_x, eth_y))
        logger.warning("In _logging_retry")
        for x, y in to_reach:
            logger.warning("Still need to reach {}:{}", x, y)
        while len(to_reach) > 0:
            just_reached = found
            found = set()
            for x, y in just_reached:
                logger.warning("Trying from {}:{}", x, y)
                # Check links starting with the most direct from 0,0
                for link_id in [1, 0, 2, 5, 3, 4]:
                    # Get protential destination
                    destination = self._machine.xy_over_link(x, y, link_id)
                    # If it is useful
                    if destination in to_reach:
                        logger.warning(
                            "Could reach {} over {}", destination, link_id)
                        # check it actually exits
                        if self._machine.is_link_at(x, y, link_id):
                            # Add to tree and record chip reachable
                            tree[destination] = (x, y, link_id)
                            to_reach.remove(destination)
                            found.add(destination)
                        else:
                            logger.error("Link down")
            logger.warning("Found {}", len(found))
            if len(found) == 0:
                raise PacmanRoutingException(
                    "Unable to do data in routing on "
                    f"{ethernet_chip.ip_address}.")
        return tree

    def _add_routing_entry(self, x, y, key, processor_id=None, link_ids=None):
        """
        Adds a routing entry on this chip, creating the table if needed.

        :param int x: chip.x
        :param int y: chip.y
        :param int key: The key to use
        :param int processor_id:
            placement.p of the monitor vertex if applicable
        :param int link_id: If of the link out if applicable
        """
        table = self._routing_tables.get_routing_table_for_chip(x, y)
        if table is None:
            table = UnCompressedMulticastRoutingTable(x, y)
            self._routing_tables.add_routing_table(table)
        if processor_id is None:
            processor_ids = []
        else:
            processor_ids = [processor_id]
        if link_ids is None:
            link_ids = []
        entry = MulticastRoutingEntry(
            routing_entry_key=key, mask=ROUTING_MASK,
            processor_ids=processor_ids, link_ids=link_ids, defaultable=False)
        table.add_multicast_routing_entry(entry)

    def _add_routing_entries(self, ethernet_chip, tree):
        """
        Adds the routing entries based on the tree.

        For every chip with this Ethernet-enabled chip on the board:
            - A key is generated (and saved) for this chip.
            - A local route to the monitor core is added.
            - The tree is walked adding a route on each source to get here

        :param ~spinn_machine.Chip ethernet_chip:
            the Ethernet-enabled chip to make entries for
        :param dict(tuple(int,int),tuple(int,int,int)) tree:
            map of chips and links
        """
        eth_x = ethernet_chip.x
        eth_y = ethernet_chip.y
        key = KEY_START_VALUE
        for (x, y) in self._machine.get_existing_xys_by_ethernet(
                eth_x, eth_y):
            self._key_to_destination_map[x, y] = key
            placement = FecDataView.get_placement_of_vertex(
                FecDataView.get_monitor_by_xy(x, y))
            self._add_routing_entry(x, y, key, processor_id=placement.p)
            while (x, y) in tree:
                x, y, link = tree[(x, y)]
                self._add_routing_entry(x, y, key, link_ids=[link])
            key += N_KEYS_PER_PARTITION_ID

        # accum links to make a broad cast
        links_per_chip = defaultdict(list)
        for chip_key in tree:
            x, y, link = tree[chip_key]
            links_per_chip[x, y].append(link)

        # add broadcast router timeout keys
        time_out_key = key
        for (x, y) in self._machine.get_existing_xys_by_ethernet(
                eth_x, eth_y):
            placement = FecDataView.get_placement_of_vertex(
                FecDataView.get_monitor_by_xy(x, y))
            self._add_routing_entry(
                x, y, time_out_key, processor_id=placement.p,
                link_ids=links_per_chip[x, y])
            # update tracker
            self._time_out_keys_by_board[(eth_x, eth_y)] = key
